"""
[1] Gene H. Golub and John H. Welsch,
    Calculation of Gauss Quadrature Rules,
    Mathematics of Computation,
    Vol. 23, No. 106 (Apr., 1969), pp. 221-230+s1-s10,
    <https://doi.org/10.1090/S0025-5718-69-99647-1>,
    <https://www.semanticscholar.org/paper/Calculation-of-Gauss-Quadrature-Rules-%2A-By-Gene-Golub-Welsch/c715119d5464f614fd8ec590b732ccfea53e72c4?p2df>.

[2] W. Gautschi,
    Algorithm 726: ORTHPOL–a package of routines for generating orthogonal polynomials
    and Gauss-type quadrature rules,
    ACM Transactions on Mathematical Software (TOMS),
    Volume 20, Issue 1, March 1994,
    Pages 21-62,
    <https://doi.org/10.1145/174603.174605>,
    <https://www.cs.purdue.edu/archives/2002/wxg/codes/gauss.m>,

[3] W. Gautschi,
    How and how not to check Gaussian quadrature formulae,
    BIT Numerical Mathematics,
    June 1983, Volume 23, Issue 2, pp 209–216,
    <https://doi.org/10.1007/BF02218441>.

[4] D. Boley and G.H. Golub,
    A survey of matrix inverse eigenvalue problems,
    Inverse Problems, 1987, Volume 3, Number 4,
    <https://doi.org/10.1088/0266-5611/3/4/010>.
"""
from __future__ import annotations

import math

import numpy as np
import sympy
from mpmath import mp
from mpmath.matrices.eigen_symmetric import tridiag_eigen
from scipy.linalg import eigh_tridiagonal
from scipy.linalg.lapack import get_lapack_funcs


def coefficients_from_gauss(points, weights):
    """Given the points and weights of a Gaussian quadrature rule, this method
    reconstructs the recurrence coefficients alpha, beta as appearing in the tridiagonal
    Jacobi matrix tri(b, a, b). This is using "Method 2--orthogonal reduction" from
    (section 3.2 in [4]).  The complexity is O(n^3); a faster method is suggested in 3.3
    in [4].
    """
    n = len(points)
    assert n == len(weights)

    flt = np.vectorize(float)
    points = flt(points)
    weights = flt(weights)

    A = np.zeros((n + 1, n + 1))

    # In sytrd, the _last_ row/column of Q are e, so put the values there.
    a00 = 1.0
    A[n, n] = a00
    k = np.arange(n)
    A[k, k] = points
    A[n, :-1] = np.sqrt(weights)
    A[:-1, n] = np.sqrt(weights)

    # Implemented in
    # <https://github.com/scipy/scipy/issues/7775>
    sytrd, sytrd_lwork = get_lapack_funcs(("sytrd", "sytrd_lwork"))

    # query lwork (optional)
    lwork, info = sytrd_lwork(n + 1)
    assert info == 0

    _, d, e, _, info = sytrd(A, lwork=lwork)
    assert info == 0

    alpha = d[:-1][::-1]
    beta = e[::-1] ** 2
    int_1 = beta[0]
    beta[0] = math.nan

    return alpha, beta, int_1


def _sympy_tridiag(a, b):
    """Creates the tridiagonal sympy matrix tridiag(b, a, b)."""
    n = len(a)
    assert n == len(b)
    A = [[0 for _ in range(n)] for _ in range(n)]
    for i in range(n):
        A[i][i] = a[i]
    for i in range(n - 1):
        A[i][i + 1] = b[i + 1]
        A[i + 1][i] = b[i + 1]
    return sympy.Matrix(A)


def scheme_from_rc(alpha, beta, int_1, mode: str | None = None):
    alpha = np.asarray(alpha)
    beta = np.asarray(beta)

    if mode is None:
        # try and guess the mode
        if alpha.dtype in [np.float32, np.float64]:
            mode = "numpy"
        else:
            raise ValueError(
                'Please specify the `mode` ("sympy", "numpy", or "mpmath").'
            )

    fun = {
        "sympy": _scheme_from_rc_sympy,
        "numpy": _scheme_from_rc_numpy,
        "mpmath": _scheme_from_rc_mpmath,
    }[mode]

    return fun(alpha, beta, int_1)


# Compute the Gauss nodes and weights from the recurrence coefficients associated with a
# set of orthogonal polynomials. See [2] and
# <http://www.scientificpython.net/pyblog/radau-quadrature>.
def _scheme_from_rc_sympy(alpha, beta, int_1):
    beta[0] = int_1

    # Construct the triadiagonal matrix [sqrt(beta), alpha, sqrt(beta)]
    A = _sympy_tridiag(alpha, [sympy.sqrt(bta) for bta in beta])

    # Extract points and weights from eigenproblem
    x = []
    w = []
    for item in A.eigenvects():
        val, multiplicity, vec = item
        assert multiplicity == 1
        assert len(vec) == 1
        vec = vec[0]
        x.append(val)
        norm2 = sum(v ** 2 for v in vec)
        # simplifiction takes really long
        # w.append(sympy.simplify(beta[0] * vec[0]**2 / norm2))
        w.append(beta[0] * vec[0] ** 2 / norm2)
    # sort by x
    order = sorted(range(len(x)), key=lambda i: x[i])
    x = [x[i] for i in order]
    w = [w[i] for i in order]
    return x, w


def _scheme_from_rc_mpmath(alpha, beta, int_1):
    # Create vector
    # cut off the first value of beta (None)
    n = len(alpha)
    b = mp.zeros(n, 1)
    for i in range(n - 1):
        b[i] = mp.sqrt(beta[i + 1])

    z = mp.zeros(1, n)
    z[0, 0] = 1
    d = mp.matrix(alpha)
    tridiag_eigen(mp, d, b, z)

    # nx1 matrix -> list of mpf
    x = np.array([mp.mpf(sympy.N(xx, mp.dps)) for xx in d])
    w = np.array([mp.mpf(sympy.N(int_1, mp.dps)) * mp.power(ww, 2) for ww in z])
    return x, w


def _scheme_from_rc_numpy(alpha, beta, int_1):
    alpha = alpha.astype(np.float64)
    beta = beta.astype(np.float64)
    x, V = eigh_tridiagonal(alpha, np.sqrt(beta[1:]))
    w = int_1 * V[0, :] ** 2
    return x, w



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, fsd, untangle, z
from ._helpers import SnScheme

source = article(
    authors=["L.N. Dobrodeev"],
    title="Cubature formulas of the seventh order of accuracy for a hypersphere and a hypercube",
    journal="USSR Computational Mathematics and Mathematical Physics",
    volume="10",
    number="1",
    year="1970",
    pages="252–253",
    url="https://doi.org/10.1016/0041-5553%2870%2990084-4",
)


def dobrodeev_1970(n):
    A = frac(1, 8)
    B = frac(5 - n, 4)
    C = frac((6 - n) * (1 - n ** 2) + 36, 4 * (n + 3))
    D = frac(81, (n + 3) * (n + 6) ** 2)
    E = frac(45 * n ** 2 + 324 * n + 216, n ** 2 + 12 * n + 36) - frac(
        n * (n ** 2 - 12 * n + 65), 6
    )

    r = sqrt(frac(3, n + 6))
    data = [
        (A, fsd(n, (r, 3))),
        (B, fsd(n, (r, 2))),
        (C, fsd(n, (r, 1))),
        (D, fsd(n, (1, 1))),
        (E, z(n)),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)

    weights /= frac(27 * (n + 2) * (n + 4), (n + 6) ** 2)
    return SnScheme("Dobrodeev 1970", n, weights, points, 7, source)



import math

import sympy


def volume(n, alpha=0, symbolic=False):
    pi = sympy.pi if symbolic else math.pi
    gamma = sympy.gamma if symbolic else math.gamma
    n = sympy.S(n) if symbolic else n

    def _recurrence(n):
        assert n >= 0
        if n == 1:
            return 2 * gamma(1 + alpha)
        elif n == 2:
            return 2 * pi * gamma(2 + alpha)
        return _recurrence(n - 2) * 2 * pi * (n + alpha - 1) * (n + alpha - 2) / (n - 2)

    return _recurrence(n)


def integrate_monomial(exponents, alpha=0, symbolic=False):
    S = sympy.S if symbolic else lambda x: x
    exponents = list(exponents)
    n = S(len(exponents))

    if any(k % 2 == 1 for k in exponents):
        return 0

    def _recurrence(exponents):
        if all(k == 0 for k in exponents):
            return volume(n, alpha, symbolic)

        # find first nonzero
        idx, k0 = next((i, k) for i, k in enumerate(exponents) if k > 0)
        p = sum(exponents)
        k2 = exponents.copy()
        k2[idx] -= 2
        return (
            _recurrence(k2)
            * (alpha + n + p - 1)
            * (alpha + p + n - 2)
            * (k0 - 1)
            / (n + p - 2)
        )

    return _recurrence(exponents)



import pathlib

from ...helpers import article
from .._helpers import _read, register

_source = article(
    authors=["F.D. Witherden", "P.E. Vincent"],
    title="On the identification of symmetric quadrature rules for finite element methods",
    journal="Computers & Mathematics with Applications",
    volume="69",
    number="10",
    month="may",
    year="2015",
    pages="1232–1241",
    url="https://doi.org/10.1016/j.camwa.2015.03.017",
)

this_dir = pathlib.Path(__file__).resolve().parent


def witherden_vincent_01():
    return _read(this_dir / "witherden_vincent_01.json", _source)


def witherden_vincent_03():
    return _read(this_dir / "witherden_vincent_03.json", _source)


def witherden_vincent_05():
    return _read(this_dir / "witherden_vincent_05.json", _source)


def witherden_vincent_07():
    return _read(this_dir / "witherden_vincent_07.json", _source)


def witherden_vincent_09():
    return _read(this_dir / "witherden_vincent_09.json", _source)


def witherden_vincent_11():
    return _read(this_dir / "witherden_vincent_11.json", _source)


def witherden_vincent_13():
    return _read(this_dir / "witherden_vincent_13.json", _source)


def witherden_vincent_15():
    return _read(this_dir / "witherden_vincent_15.json", _source)


def witherden_vincent_17():
    return _read(this_dir / "witherden_vincent_17.json", _source)


def witherden_vincent_19():
    return _read(this_dir / "witherden_vincent_19.json", _source)


def witherden_vincent_21():
    return _read(this_dir / "witherden_vincent_21.json", _source)


register(
    [
        witherden_vincent_01,
        witherden_vincent_03,
        witherden_vincent_05,
        witherden_vincent_07,
        witherden_vincent_09,
        witherden_vincent_11,
        witherden_vincent_13,
        witherden_vincent_15,
        witherden_vincent_17,
        witherden_vincent_19,
        witherden_vincent_21,
    ]
)



import math

import numpy as np
import sympy

from .. import un
from ..helpers import article, fsd, pm, untangle
from ._helpers import SnScheme

source = article(
    authors=["A.H. Stroud"],
    title="Some Seventh Degree Integration Formulas for Symmetric Regions",
    journal="SIAM J. Numer. Anal.",
    volume="4",
    number="1",
    pages="37–44",
    url="https://doi.org/10.1137/0704004",
)


def _stroud_1967_7_ab(n, variant_a, symbolic):
    sqrt = sympy.sqrt if symbolic else math.sqrt

    if variant_a:
        assert 3 <= n <= 7
        t = 1
    else:
        # ERR Stroud mentions nothing of variant b being only valid up to dimension 6,
        # but that's the way it is.
        assert 3 <= n <= 6
        t = -1

    alpha = sqrt(6 * (n + 6) * (8 - n))

    r2 = (3 * (n + 6) * (8 - n) - t * (n - 2) * alpha) / ((n + 6) * (34 - 5 * n))
    s2 = (3 * n * (n + 6) - t * 2 * alpha) / ((n + 6) * (3 * n ** 2 + 6 * n - 16))
    t2 = (6 * (n + 6) + t * alpha) / (14 * (n + 6))

    B = (8 - n) / (n + 2) / (n + 4) / (n + 6) / r2 ** 3
    C = 1 / (n + 2) / (n + 4) / (n + 6) / s2 ** 3 / 2 ** n
    D = 1 / (n + 2) / (n + 4) / (n + 6) / t2 ** 3 / 2
    A = 1 - 2 * n * B - 2 ** n * C - 2 * n * (n - 1) * D

    r = sqrt(r2)
    s = sqrt(s2)
    t = sqrt(t2)

    data = [(A, [n * [0]]), (B, fsd(n, (r, 1))), (C, pm(n * [s])), (D, fsd(n, (t, 2)))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)

    variant = "a" if variant_a else "b"
    return SnScheme(f"Stroud 1967-7{variant}", n, weights, points, 7, source)


def stroud_1967_7_a(n, symbolic=False):
    return _stroud_1967_7_ab(n, variant_a=True, symbolic=symbolic)


def stroud_1967_7_b(n, symbolic=False):
    return _stroud_1967_7_ab(n, variant_a=False, symbolic=symbolic)


def stroud_1967_7_c(n, symbolic=False):
    sqrt = sympy.sqrt if symbolic else math.sqrt
    assert n >= 3

    alpha = sqrt(2 * (n + 2) * (n + 4))

    r1, r2 = (
        sqrt(((n + 2) * (n + 4) + i * 2 * alpha) / (n + 4) / (n + 6)) for i in [+1, -1]
    )
    A1, A2 = (
        (2 * (n + 2) ** 2 + i * (n - 2) * alpha) / (4 * n * (n + 2) ** 2)
        for i in [+1, -1]
    )

    s = un.stroud_1967(n)

    points = np.concatenate([r1 * s.points, r2 * s.points])
    points = np.ascontiguousarray(points.T)
    weights = np.concatenate([A1 * s.weights, A2 * s.weights])

    # weights *= un.volume_nsphere(n - 1, symbolic) / volume_nball(n, symbolic)
    weights *= n

    return SnScheme("Stroud 1967-7 c", n, weights, points, 7, source)



from typing import Callable

import numpy as np

from .c1 import integrate_adaptive


# compatibility for scipy.quad
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.quad.html
def quad(
    f: Callable,
    a,
    b,
    args=(),
    epsabs: float = 1.49e-08,
    epsrel: float = 1.49e-08,
    limit: int = 50,
):
    # See <https://www.gnu.org/software/gsl/doc/html/integration.html> for the
    # variable transformations
    if a == -np.inf and b == np.inf:
        # x = (1 - t) / t
        # dx / dt = -1 / t**2
        a = 0.0
        b = 1.0

        def g(t):
            return (f((1 - t) / t, *args) + f(-(1 - t) / t)) / t ** 2

    elif b == np.inf:
        a_orig = a
        a = 0.0
        b = 1.0

        def g(t):
            return f(a_orig + (1 - t) / t, *args) / t ** 2

    elif a == -np.inf:
        b_orig = b
        a = 0.0
        b = 1.0

        def g(t):
            return f(b_orig - (1 - t) / t, *args) / t ** 2

    else:

        def g(x):
            return f(x, *args)

    swap = a > b
    if swap:
        a, b = b, a

    val, err = integrate_adaptive(
        g,
        [a, b],
        eps_abs=epsabs,
        eps_rel=epsrel,
        criteria_connection=np.any,
        max_num_subintervals=limit,
    )
    if swap:
        val *= -1

    return val, err



import numpy as np
import sympy

from ..helpers import article, get_all_exponents, prod
from ._helpers import TnScheme

source = article(
    authors=["P. Silvester"],
    title="Symmetric quadrature formulae for simplexes",
    journal="Math. Comp.",
    volume="24",
    pages="95-100",
    year="1970",
    url="https://doi.org/10.1090/S0025-5718-1970-0258283-6",
)


def integrate_bary(k, symbolic):
    # This is very similar to integrate_monomial_over_unit_simplex, except that n =
    # len(k) - 1, not len(k). This is because we're integrating over a polynomial in
    # barycentric coordinates here. Also the integration for [0, ..., 0] is set equal to
    # 1. This ensures that the weights add up to 1, not the volume of the unit simplex.
    frac = sympy.Rational if symbolic else lambda a, b: a / b

    n = len(k) - 1
    assert all(kk >= 0 for kk in k)

    if all(kk == 0 for kk in k):
        # return frac(1, math.factorial(n))
        return 1

    # find first nonzero
    idx = next(i for i, j in enumerate(k) if j > 0)
    alpha = frac(k[idx], sum(k) + n)
    k2 = k.copy()
    k2[idx] -= 1
    return integrate_bary(k2, symbolic) * alpha


def _get_data(dim, n, point_fun, symbolic):
    # points
    idxs = np.array(get_all_exponents(dim + 1, n)[-1])
    points = point_fun(idxs)

    # weights
    weights = np.empty(len(points))
    kk = 0
    for idx in idxs:
        # Define the polynomial which to integrate over the simplex.
        t = sympy.DeferredVector("t")
        g = prod(
            sympy.poly((t[i] - point_fun(k)) / (point_fun(m) - point_fun(k)))
            for i, m in enumerate(idx)
            for k in range(m)
        )
        if isinstance(g, int):
            exp = [0] * (dim + 1)
            coeffs = g
        else:
            # make sure the exponents are all of the correct length
            exp = [list(m) + [0] * (dim - len(m) + 1) for m in g.monoms()]
            coeffs = g.coeffs()
        weights[kk] = sum(c * integrate_bary(m, symbolic) for c, m in zip(coeffs, exp))
        kk += 1

    return weights, points


def silvester(dim, variant, n, symbolic=False):
    frac = np.vectorize(sympy.Rational) if symbolic else lambda a, b: a / b

    def points1d_closed(k):
        return frac(k, n)

    def points1d_open(k):
        return frac(k + 1, n + 1 + dim)

    if variant == "closed":
        degree = n
        tol = 1.034e-14
        points1d = points1d_closed
    else:
        assert variant == "open"
        degree = 1 if n == 0 else n
        tol = 1.040e-13
        points1d = points1d_open

    weights, points = _get_data(dim, n, points1d, symbolic)
    points = np.ascontiguousarray(points.T)

    return TnScheme(
        f"Silvester ({variant}, dim={n})", dim, weights, points, degree, source, tol
    )



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["B. Vioreanu", "V. Rokhlin"],
    title="Spectra of Multiplication Operators as a Numerical Tool",
    year="2014",
    journal="SIAM J. Sci. Comput.",
    volume="36",
    number="1",
    url="https://doi.org/10.1137/110860082",
)
# http://www.cs.yale.edu/publications/techreports/tr1443.pdf

# Data adapted from modepy
# https://github.com/inducer/modepy/blob/master/modepy/quadrature/vr_quad_data_tri.py

this_dir = pathlib.Path(__file__).resolve().parent


def vioreanu_rokhlin_00():
    return _read(this_dir / "vr00.json", source)


def vioreanu_rokhlin_01():
    return _read(this_dir / "vr01.json", source)


def vioreanu_rokhlin_02():
    return _read(this_dir / "vr02.json", source)


def vioreanu_rokhlin_03():
    return _read(this_dir / "vr03.json", source)


def vioreanu_rokhlin_04():
    return _read(this_dir / "vr04.json", source)


def vioreanu_rokhlin_05():
    return _read(this_dir / "vr05.json", source)


def vioreanu_rokhlin_06():
    return _read(this_dir / "vr06.json", source)


def vioreanu_rokhlin_07():
    return _read(this_dir / "vr07.json", source)


def vioreanu_rokhlin_08():
    return _read(this_dir / "vr08.json", source)


def vioreanu_rokhlin_09():
    return _read(this_dir / "vr09.json", source)


def vioreanu_rokhlin_10():
    return _read(this_dir / "vr10.json", source)


def vioreanu_rokhlin_11():
    return _read(this_dir / "vr11.json", source)


def vioreanu_rokhlin_12():
    return _read(this_dir / "vr12.json", source)


def vioreanu_rokhlin_13():
    return _read(this_dir / "vr13.json", source)


def vioreanu_rokhlin_14():
    return _read(this_dir / "vr14.json", source)


def vioreanu_rokhlin_15():
    return _read(this_dir / "vr15.json", source)


def vioreanu_rokhlin_16():
    return _read(this_dir / "vr16.json", source)


def vioreanu_rokhlin_17():
    return _read(this_dir / "vr17.json", source)


def vioreanu_rokhlin_18():
    return _read(this_dir / "vr18.json", source)


def vioreanu_rokhlin_19():
    return _read(this_dir / "vr19.json", source)


register(
    [
        vioreanu_rokhlin_00,
        vioreanu_rokhlin_01,
        vioreanu_rokhlin_02,
        vioreanu_rokhlin_03,
        vioreanu_rokhlin_04,
        vioreanu_rokhlin_05,
        vioreanu_rokhlin_06,
        vioreanu_rokhlin_07,
        vioreanu_rokhlin_08,
        vioreanu_rokhlin_09,
        vioreanu_rokhlin_10,
        vioreanu_rokhlin_10,
        vioreanu_rokhlin_11,
        vioreanu_rokhlin_12,
        vioreanu_rokhlin_13,
        vioreanu_rokhlin_14,
        vioreanu_rokhlin_15,
        vioreanu_rokhlin_16,
        vioreanu_rokhlin_17,
        vioreanu_rokhlin_18,
        vioreanu_rokhlin_19,
    ]
)



import math

import sympy


def volume_physicists(n, symbolic):
    pi = sympy.pi if symbolic else math.pi
    sqrt = sympy.sqrt if symbolic else math.sqrt

    def _recurrence(n):
        assert n >= 0
        if n == 0:
            return 1
        elif n == 1:
            return sqrt(pi)
        return _recurrence(n - 2) * pi

    return _recurrence(n)


def integrate_monomial_physicists(exponents, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    exponents = list(exponents)

    if any(k % 2 == 1 for k in exponents):
        return 0

    def _recurrence(exponents):
        if all(k == 0 for k in exponents):
            n = len(exponents)
            return volume_physicists(n, symbolic)

        # find first nonzero
        idx, k0 = next((i, k) for i, k in enumerate(exponents) if k > 0)
        k2 = exponents.copy()
        k2[idx] -= 2
        return _recurrence(k2) * frac(k0 - 1, 2)

    return _recurrence(exponents)


def volume_probabilists(n):
    return 1


def integrate_monomial_probabilists(exponents):
    exponents = list(exponents)

    if any(k % 2 == 1 for k in exponents):
        return 0

    def _recurrence(exponents):
        if all(k == 0 for k in exponents):
            n = len(exponents)
            return volume_probabilists(n)

        # find first nonzero
        idx, k0 = next((i, k) for i, k in enumerate(exponents) if k > 0)
        k2 = exponents.copy()
        k2[idx] -= 2
        return _recurrence(k2) * (k0 - 1)

    return _recurrence(exponents)


def volume(n, variant, symbolic=False):
    if variant == "physicists":
        return volume_physicists(n, symbolic)
    assert variant == "probabilists"
    return volume_probabilists(n)


def integrate_monomial(exponents, variant, symbolic=False):
    if variant == "physicists":
        return integrate_monomial_physicists(exponents, symbolic)
    assert variant == "probabilists"
    return integrate_monomial_probabilists(exponents)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["S. Wandzura", "H. Xiao"],
    title="Symmetric quadrature rules on a triangle",
    journal="Computers & Mathematics with Applications",
    volume="45",
    number="12",
    month="jun",
    year="2003",
    pages="1829-1840",
    url="https://doi.org/10.1016/S0898-1221%2803%2990004-6",
)
# Note that in the above article, the authors present the coordinates in the symmetric
# triangle [[-0.5, -sqrt(3)/2], [-0.5, +sqrt(3)/2], [1, 0]]. These have been transformed
# to barycentric coordinates here.

this_dir = pathlib.Path(__file__).resolve().parent


def wandzura_xiao_1():
    return _read(this_dir / "wx01.json", source)


def wandzura_xiao_2():
    return _read(this_dir / "wx02.json", source)


def wandzura_xiao_3():
    return _read(this_dir / "wx03.json", source)


def wandzura_xiao_4():
    return _read(this_dir / "wx04.json", source)


def wandzura_xiao_5():
    return _read(this_dir / "wx05.json", source)


def wandzura_xiao_6():
    return _read(this_dir / "wx06.json", source)


register(
    [
        wandzura_xiao_1,
        wandzura_xiao_2,
        wandzura_xiao_3,
        wandzura_xiao_4,
        wandzura_xiao_5,
        wandzura_xiao_6,
    ]
)



from ._dobrodeev_1970 import dobrodeev_1970
from ._dobrodeev_1978 import dobrodeev_1978
from ._hammer_stroud import hammer_stroud_11n, hammer_stroud_12n
from ._mcnamee_stenger import (
    mcnamee_stenger_3,
    mcnamee_stenger_5,
    mcnamee_stenger_7a,
    mcnamee_stenger_7b,
    mcnamee_stenger_9a,
    mcnamee_stenger_9b,
)
from ._stoyanova import stoyanova
from ._stroud import stroud_sn_3_1  # stroud_sn_7_3a,
from ._stroud import (
    stroud_sn_2_1,
    stroud_sn_3_2,
    stroud_sn_5_1a,
    stroud_sn_5_1b,
    stroud_sn_5_2,
    stroud_sn_5_3,
    stroud_sn_5_4,
    stroud_sn_5_5,
    stroud_sn_5_6,
    stroud_sn_7_1a,
    stroud_sn_7_1b,
    stroud_sn_7_2,
    stroud_sn_7_3a,
    stroud_sn_7_3b,
    stroud_sn_9_1a,
    stroud_sn_9_1b,
    stroud_sn_11_1a,
    stroud_sn_11_1b,
)
from ._stroud_1957 import stroud_1957
from ._stroud_1966 import stroud_1966_a, stroud_1966_b, stroud_1966_c, stroud_1966_d
from ._stroud_1967_5 import stroud_1967_5_a, stroud_1967_5_b
from ._stroud_1967_7 import stroud_1967_7_a, stroud_1967_7_b, stroud_1967_7_c

__all__ = [
    "dobrodeev_1970",
    "dobrodeev_1978",
    "hammer_stroud_11n",
    "hammer_stroud_12n",
    "mcnamee_stenger_3",
    "mcnamee_stenger_5",
    "mcnamee_stenger_7a",
    "mcnamee_stenger_7b",
    "mcnamee_stenger_9a",
    "mcnamee_stenger_9b",
    "stoyanova",
    "stroud_sn_2_1",
    "stroud_sn_3_1",
    "stroud_sn_3_2",
    "stroud_sn_5_1a",
    "stroud_sn_5_1b",
    "stroud_sn_5_2",
    "stroud_sn_5_3",
    "stroud_sn_5_4",
    "stroud_sn_5_5",
    "stroud_sn_5_6",
    "stroud_sn_7_1a",
    "stroud_sn_7_1b",
    "stroud_sn_7_2",
    "stroud_sn_7_3a",
    "stroud_sn_7_3b",
    "stroud_sn_9_1a",
    "stroud_sn_9_1b",
    "stroud_sn_11_1a",
    "stroud_sn_11_1b",
    "stroud_1957",
    "stroud_1966_a",
    "stroud_1966_b",
    "stroud_1966_c",
    "stroud_1966_d",
    "stroud_1967_5_a",
    "stroud_1967_5_b",
    "stroud_1967_7_a",
    "stroud_1967_7_b",
    "stroud_1967_7_c",
]



import numpy as np
from sympy import Rational as frac

from ..helpers import article, rd, untangle
from ._helpers import TnScheme

source = article(
    authors=["A.H. Stroud"],
    title="Numerical Integration Formulas of Degree 3 for Product Regions and Cones",
    journal="Mathematics of Computation",
    volume="15",
    number="74",
    month="apr",
    year="1961",
    pages="143-150",
    url="https://doi.org/10.2307/2004220",
)


def stroud_1961(n):
    degree = 3

    r = frac(1, n + 1)
    s = frac(1, n)

    prod = (n + 1) * (n + 2) * (n + 3)
    A = frac((3 - n) * (n + 1) ** 3, prod)
    B = frac(3, prod)
    C = frac(n ** 3, prod)

    data = [(A, [(n + 1) * [r]]), (B, rd(n + 1, [(1, 1)])), (C, rd(n + 1, [(s, n)]))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud 1961", n, weights, points, degree, source)



import numpy as np

from ..helpers import article
from ._helpers import C1Scheme

source = article(
    authors=["J. Waldvogel"],
    title="Fast Construction of the Fejér and Clenshaw–Curtis Quadrature Rules",
    journal="BIT Numerical Mathematics",
    month="mar",
    year="2006",
    volume="46",
    number="1",
    pages="195–202",
    url="https://doi.org/10.1007/s10543-006-0045-4",
)


def fejer_1(n):
    degree = n

    points = -np.cos(np.pi * (np.arange(n) + 0.5) / n)

    # n -= 1
    N = np.arange(1, n, 2)
    length = len(N)
    m = n - length
    K = np.arange(m)

    v0 = np.concatenate(
        [
            2 * np.exp(1j * np.pi * K / n) / (1 - 4 * K ** 2),
            np.zeros(length + 1),
        ]
    )
    v1 = v0[:-1] + np.conjugate(v0[:0:-1])

    w = np.fft.ifft(v1)
    assert max(w.imag) < 1.0e-15
    weights = w.real

    return C1Scheme("Fejér 1", degree, weights, points, source)


def fejer_2(n):
    degree = n

    points = -np.cos((np.pi * np.arange(1, n + 1)) / (n + 1))

    n += 1
    N = np.arange(1, n, 2)
    length = len(N)
    m = n - length
    v0 = np.concatenate([2.0 / N / (N - 2), np.array([1.0 / N[-1]]), np.zeros(m)])
    v2 = -v0[:-1] - v0[:0:-1]

    w = np.fft.ihfft(v2)
    assert max(w.imag) < 1.0e-15
    w = w.real

    if n % 2 == 1:
        weights = np.concatenate([w, w[::-1]])
    else:
        weights = np.concatenate([w, w[len(w) - 2 :: -1]])

    # cut off first and last
    weights = weights[1:-1]
    return C1Scheme("Fejér 2", degree, weights, points, source)



import pathlib

import numpy as np
from sympy import Rational as frac

from ...helpers import article, book
from .._hammer_marlowe_stroud import hammer_marlowe_stroud_3 as strang_fix_cowper_02
from .._helpers import T2Scheme, _read, register

source = book(
    authors=["Gilbert Strang", "George Fix"],
    title="An Analysis of the Finite Element Method",
    publisher="Wellesley-Cambridge Press",
    year="1973",
    isbn="096140888X",
    url="https://bookstore.siam.org/wc08/",
)

source = article(
    authors=["G.R. Cowper"],
    title="Gaussian quadrature formulas for triangles",
    journal="Numerical Methods in Engineering",
    volume="7",
    number="3",
    year="1973",
    pages="405–408",
    url="https://doi.org/10.1002/nme.1620070316",
)
# See also
# https://people.sc.fsu.edu/~jburkardt/datasets/quadrature_rules_tri/quadrature_rules_tri.html

this_dir = pathlib.Path(__file__).resolve().parent


def strang_fix_cowper_01():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 6)]]}
    return T2Scheme("Strang-Fix-Cowper 1", d, 2, source)


def strang_fix_cowper_03():
    d = {"centroid": [[-frac(9, 16)]], "d3_aa": [[frac(25, 48)], [frac(1, 5)]]}
    return T2Scheme("Strang-Fix-Cowper 3", d, 3, source)


def strang_fix_cowper_04():
    roots = np.polynomial.polynomial.polyroots([-1, 15, -60, 60])
    d = {"d3_ab": [[1 / 6], [roots[2]], [roots[1]]]}
    return T2Scheme("Strang-Fix-Cowper 4", d, 3, source, 2.463e-15)


def strang_fix_cowper_05():
    return _read(this_dir / "strang_fix_cowper_05.json", source)


def strang_fix_cowper_06():
    d = {
        "centroid": [[3 / 8]],
        "d3_ab": [[5 / 48], [0.736712498968435], [0.237932366472434]],
    }
    return T2Scheme("Strang-Fix-Cowper 6", d, 4, source, 5.440e-15)


def strang_fix_cowper_07():
    d = {
        "centroid": [[9 / 40]],
        "d3_aa": [
            [0.12593918054482717, 0.13239415278850616],
            [0.10128650732345633, 0.47014206410511505],
        ],
    }
    return T2Scheme("Strang-Fix-Cowper 7", d, 5, source, 5.551e-16)


def strang_fix_cowper_08():
    return _read(this_dir / "strang_fix_cowper_08.json", source)


def strang_fix_cowper_09():
    return _read(this_dir / "strang_fix_cowper_09.json", source)


def strang_fix_cowper_10():
    return _read(this_dir / "strang_fix_cowper_10.json", source)


register(
    [
        strang_fix_cowper_01,
        strang_fix_cowper_02,
        strang_fix_cowper_03,
        strang_fix_cowper_04,
        strang_fix_cowper_05,
        strang_fix_cowper_06,
        strang_fix_cowper_07,
        strang_fix_cowper_08,
        strang_fix_cowper_09,
        strang_fix_cowper_10,
    ]
)



# ENH quadpy-optimize
import pathlib

from ...helpers import article
from .._helpers import _read, register

_source = article(
    authors=["KyoungJoong Kim", "ManSuk Song"],
    title="Symmetric quadrature formulas over a unit disk",
    journal="Korean J. Comp. & Appl. Math.",
    year="1997",
    volume="4",
    pages="179-192",
    url="https://doi.org/10.1007/BF03011388",
)

this_dir = pathlib.Path(__file__).resolve().parent


def kim_song_1():
    return _read(this_dir / "kim_song_01.json", _source)


def kim_song_2():
    return _read(this_dir / "kim_song_02.json", _source)


def kim_song_3():
    return _read(this_dir / "kim_song_03.json", _source)


def kim_song_4():
    return _read(this_dir / "kim_song_04.json", _source)


def kim_song_5():
    return _read(this_dir / "kim_song_05.json", _source)


def kim_song_6():
    return _read(this_dir / "kim_song_06.json", _source)


def kim_song_7():
    return _read(this_dir / "kim_song_07.json", _source)


# TODO find issue
def kim_song_8():
    return _read(this_dir / "kim_song_08.json", _source)


def kim_song_9():
    return _read(this_dir / "kim_song_09.json", _source)


def kim_song_10():
    return _read(this_dir / "kim_song_10.json", _source)


def kim_song_11():
    return _read(this_dir / "kim_song_11.json", _source)


def kim_song_12():
    return _read(this_dir / "kim_song_12.json", _source)


def kim_song_13():
    return _read(this_dir / "kim_song_13.json", _source)


def kim_song_14():
    return _read(this_dir / "kim_song_14.json", _source)


def kim_song_15():
    return _read(this_dir / "kim_song_15.json", _source)


register(
    [
        kim_song_1,
        kim_song_2,
        kim_song_3,
        kim_song_4,
        kim_song_5,
        kim_song_6,
        kim_song_7,
        kim_song_8,
        kim_song_9,
        kim_song_10,
        kim_song_11,
        kim_song_12,
        kim_song_13,
        kim_song_14,
        kim_song_15,
    ]
)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["Stefanos-Aldo Papanicolopulos"],
    title="New fully symmetric and rotationally symmetric cubature rules on the triangle using minimal orthonormal bases",
    journal="Journal of Computational and Applied Mathematics",
    volume="294",
    month="mar",
    year="2016",
    pages="39–48",
    url="https://doi.org/10.1016/j.cam.2015.08.001",
)
# https://arxiv.org/abs/1411.5631

this_dir = pathlib.Path(__file__).resolve().parent


def papanicolopulos_sym_0():
    return _read(this_dir / "full00.json", source)


def papanicolopulos_sym_1():
    return _read(this_dir / "full01.json", source)


def papanicolopulos_sym_2():
    return _read(this_dir / "full01.json", source)


def papanicolopulos_sym_3():
    return _read(this_dir / "full01.json", source)


def papanicolopulos_sym_4():
    return _read(this_dir / "full01.json", source)


def papanicolopulos_sym_5():
    return _read(this_dir / "full01.json", source)


def papanicolopulos_sym_6():
    return _read(this_dir / "full01.json", source)


def papanicolopulos_sym_7():
    return _read(this_dir / "full01.json", source)


def papanicolopulos_sym_8():
    return _read(this_dir / "full08.json", source)


# TODO ERR the first 8 schemes are flawed by round-off error
# def papanicolopulos_rot_00():
#     return T2Scheme("Papanicolopulos 0 (rotational symmetry)", *_read("rot00.json"))
#
#
# def papanicolopulos_rot_01():
#     return T2Scheme("Papanicolopulos 1 (rotational symmetry)", *_read("rot01.json"))
#
#
# def papanicolopulos_rot_02():
#     return T2Scheme("Papanicolopulos 2 (rotational symmetry)", *_read("rot02.json"))
#
#
# def papanicolopulos_rot_03():
#     return T2Scheme("Papanicolopulos 3 (rotational symmetry)", *_read("rot03.json"))
#
#
# def papanicolopulos_rot_04():
#     return T2Scheme("Papanicolopulos 4 (rotational symmetry)", *_read("rot04.json"))
#
#
# def papanicolopulos_rot_05():
#     return T2Scheme("Papanicolopulos 5 (rotational symmetry)", *_read("rot05.json"))
#
#
# def papanicolopulos_rot_06():
#     return T2Scheme("Papanicolopulos 6 (rotational symmetry)", *_read("rot06.json"))
#
#
# def papanicolopulos_rot_07():
#     return T2Scheme("Papanicolopulos 7 (rotational symmetry)", *_read("rot07.json"))


def papanicolopulos_rot_08():
    return _read(this_dir / "rot08.json", source)


def papanicolopulos_rot_09():
    return _read(this_dir / "rot09.json", source)


def papanicolopulos_rot_10():
    return _read(this_dir / "rot10.json", source)


def papanicolopulos_rot_11():
    return _read(this_dir / "rot11.json", source)


def papanicolopulos_rot_12():
    return _read(this_dir / "rot12.json", source)


def papanicolopulos_rot_13():
    return _read(this_dir / "rot13.json", source)


def papanicolopulos_rot_14():
    return _read(this_dir / "rot14.json", source)


def papanicolopulos_rot_15():
    return _read(this_dir / "rot15.json", source)


def papanicolopulos_rot_16():
    return _read(this_dir / "rot16.json", source)


def papanicolopulos_rot_17():
    return _read(this_dir / "rot17.json", source)


register(
    [
        papanicolopulos_sym_0,
        papanicolopulos_sym_1,
        papanicolopulos_sym_2,
        papanicolopulos_sym_3,
        papanicolopulos_sym_4,
        papanicolopulos_sym_5,
        papanicolopulos_sym_6,
        papanicolopulos_sym_7,
        papanicolopulos_sym_8,
        #
        papanicolopulos_rot_08,
        papanicolopulos_rot_09,
        papanicolopulos_rot_10,
        papanicolopulos_rot_11,
        papanicolopulos_rot_12,
        papanicolopulos_rot_13,
        papanicolopulos_rot_14,
        papanicolopulos_rot_15,
        papanicolopulos_rot_16,
        papanicolopulos_rot_17,
    ]
)



import math

import numpy as np
import sympy

from ..helpers import article, combine, fsd, pm, untangle, z
from ._helpers import SnScheme

source = article(
    authors=["A.H. Stroud"],
    title="Some Fifth Degree Integration Formulas for Symmetric Regions",
    journal="Mathematics of Computation",
    volume="20",
    number="93",
    month="jan",
    year="1966",
    pages="90-97",
    url="https://doi.org/10.1090/S0025-5718-1966-0191094-8",
)


def stroud_1966_a(n, symbolic=False):
    sqrt = sympy.sqrt if symbolic else math.sqrt

    a = sqrt(2 * (n + 4))

    r2 = (n + 4 - a) / (n + 4)
    s2 = (n * (n + 4) + 2 * a) / (n ** 2 + 2 * n - 4) / (n + 4)

    B1 = 1 / r2 ** 2 / (n + 2) / (n + 4)
    B2 = 1 / s2 ** 2 / 2 ** n / (n + 2) / (n + 4)

    r = sqrt(r2)
    s = sqrt(s2)

    data = [(B1, fsd(n, (r, 1))), (B2, pm(n * [s]))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return SnScheme("Stroud 1966a", n, weights, points, 5, source)


def stroud_1966_b(n, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt

    alpha = 0
    s = sqrt(frac(n + alpha + 2, (n + 2) * (n + alpha + 4)))
    data = []
    B0 = 1
    for k in range(1, n + 1):
        B = frac(
            2 ** (k - n) * (n + 2) * (n + alpha) * (n + alpha + 4),
            n * (k + 1) * (k + 2) * (n + alpha + 2) ** 2,
        )
        B0 -= 2 ** (n - k + 1) * B
        r = sqrt(frac((k + 2) * (n + alpha + 2), (n + 2) * (n + alpha + 4)))
        v = np.concatenate(
            [
                np.zeros((2 ** (n - k + 1), k - 1), dtype=int),
                pm(np.array([r] + (n - k) * [s])),
            ],
            axis=-1,
        )
        data.append((B, v))
    data.append((B0, z(n)))

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return SnScheme("Stroud 1966b", n, weights, points, 5, source)


def stroud_1966_c(n, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt

    a = sqrt(2 * (n + 2))
    r = sqrt((n + 2 + (n - 1) * a) / (n * (n + 4)))
    s = sqrt((n + 2 - a) / (n * (n + 4)))

    B0 = frac(4, (n + 2) ** 2)
    B1 = frac(n + 4, 2 ** n * (n + 2) ** 2)

    data = [(B0, z(n)), (B1, combine(((+r, -r), 1), ((+s, -s), (n - 1))))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return SnScheme("Stroud 1966c", n, weights, points, 5, source)


def stroud_1966_d(n, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt

    a = 2 * sqrt(n + 4)
    b = sqrt(2 * (n + 1) * (n + 2) * (n + 4))

    r = sqrt((n * (n + 4) + a + (n - 1) * b) / (n * (n + 2) * (n + 4)))
    s = sqrt((n * (n + 4) + a - b) / (n * (n + 2) * (n + 4)))
    t = sqrt((n + 4 - a) / ((n + 2) * (n + 4)))

    B = frac(1, 2 ** n * (n + 1))

    # The data is given symbolically, and for large n, those are thousands of points and
    # weights. Converting them to float takes a long time. A better approach would be to
    # be convert r, s, t first, and assemble the data afterwards.
    data = [(B, combine(((+r, -r), 1), ((+s, -s), (n - 1)))), (B, pm(n * [t]))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return SnScheme("Stroud 1966d", n, weights, points, 5, source, 1.019e-14)



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C2Scheme, register

source = article(
    authors=["G.M. Phillips"],
    title="Numerical integration in two and three dimensions",
    journal="Comput J",
    year="1967",
    volume="10",
    number="2",
    pages="202-204",
    url="https://doi.org/10.1093/comjnl/10.2.202",
)
# TODO add scheme for hex


def phillips():
    c = 3 * sqrt(385)
    r, s = (sqrt((105 + i * c) / 140) for i in [+1, -1])
    t = sqrt(frac(3, 5))

    B1, B2 = ((77 - i * c) / 891 for i in [+1, -1])
    B3 = frac(25, 324)

    d = {
        "d4_a0": [[B1, B2], [r, s]],
        "d4_aa": [[B3], [t]],
    }
    return C2Scheme("Phillips", d, 7, source)


register([phillips])



import numpy as np

from ..c1 import gauss_legendre
from ..helpers import book
from ._albrecht_collatz import albrecht_collatz as stroud_t2_3_1
from ._hammer_marlowe_stroud import hammer_marlowe_stroud_5 as stroud_t2_5_1
from ._helpers import T2Scheme, register

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_t2_7_1():
    # conical product Gauss
    gl4 = gauss_legendre(4)
    r = (gl4.points + 1) / 2
    A = gl4.weights / 2

    # Generate Gauss formula for int_0^1 (1-s) * f(s) ds.
    # ```
    # k = np.arange(8)
    # moments = 1 / (k**2 + 3*k + 2)
    # alpha, beta = orthopy.line.chebyshev(moments)
    # s, B = orthopy.line.schemes.custom(alpha, beta, mode='numpy')
    # ```
    s = np.array(
        [
            5.710419611452533e-02,
            2.768430136381415e-01,
            5.835904323689318e-01,
            8.602401356562251e-01,
        ]
    )
    B = np.array(
        [
            1.355069134315012e-01,
            2.034645680102685e-01,
            1.298475476082247e-01,
            3.118097095000554e-02,
        ]
    )

    weights = np.array([2 * A[i] * B[j] for i in range(4) for j in range(4)])
    points = np.array(
        [
            [s[j], r[i] * (1 - s[j]), (1 - r[i]) * (1 - s[j])]
            for i in range(4)
            for j in range(4)
        ]
    )
    points = np.ascontiguousarray(points.T)
    return T2Scheme(
        "Stroud 7-1", {"plain": np.vstack([weights, points])}, 7, source, 1.403e-13
    )


register([stroud_t2_3_1, stroud_t2_5_1, stroud_t2_7_1])



from ._adaptive import integrate_adaptive
from ._chebyshev_gauss import chebyshev_gauss_1, chebyshev_gauss_2
from ._clenshaw_curtis import clenshaw_curtis
from ._fejer import fejer_1, fejer_2
from ._gauss_jacobi import gauss_jacobi
from ._gauss_kronrod import gauss_kronrod
from ._gauss_legendre import gauss_legendre
from ._gauss_lobatto import gauss_lobatto
from ._gauss_patterson import gauss_patterson
from ._gauss_radau import gauss_radau
from ._midpoint import midpoint
from ._newton_cotes import newton_cotes_closed, newton_cotes_open
from ._trapezoidal import trapezoidal

__all__ = [
    "chebyshev_gauss_1",
    "chebyshev_gauss_2",
    "clenshaw_curtis",
    "fejer_1",
    "fejer_2",
    "gauss_jacobi",
    "gauss_kronrod",
    "gauss_legendre",
    "gauss_lobatto",
    "gauss_patterson",
    "gauss_radau",
    "midpoint",
    "newton_cotes_open",
    "newton_cotes_closed",
    "trapezoidal",
    "integrate_adaptive",
]



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C2Scheme, register

source = article(
    authors=["G.W. Tyler"],
    title="Numerical integration of functions of several variables",
    journal="Canad. J. Math.",
    volume="5",
    year="1953",
    pages="393-412",
    url="https://doi.org/10.4153/CJM-1953-044-1",
)


def tyler_1():
    d = {
        "zero2": [[-frac(28, 45)]],
        "d4_aa": [[frac(1, 36)], [1]],
        "d4_a0": [[frac(1, 45), frac(16, 45)], [1, frac(1, 2)]],
    }
    return C2Scheme("Tyler 1", d, 5, source)


def tyler_2():
    r = sqrt(frac(6, 7))
    s, t = (sqrt((114 - i * 3 * sqrt(583)) / 287) for i in [+1, -1])
    B1 = frac(49, 810)
    B2, B3 = ((178981 + i * 2769 * sqrt(583)) / 1888920 for i in [+1, -1])
    d = {"d4_a0": [[B1], [r]], "d4_aa": [[B2, B3], [s, t]]}
    return C2Scheme("Tyler 2", d, 7, source, 3.331e-16)


def tyler_3():
    d = {
        "zero2": [[frac(449, 315)]],
        "d4_a0": [
            [frac(37, 1260), frac(3, 28), -frac(69, 140)],
            [1, frac(2, 3), frac(1, 3)],
        ],
        "d4_aa": [[frac(7, 540), frac(32, 135)], [1, frac(1, 2)]],
    }
    return C2Scheme("Tyler 3", d, 7, source)


register([tyler_1, tyler_2, tyler_3])



import numpy as np
import sympy

from ..helpers import book, untangle, z
from ._albrecht import albrecht_4 as stroud_s2_9_1
from ._albrecht import albrecht_5 as stroud_s2_11_2
from ._albrecht import albrecht_6 as stroud_s2_13_2
from ._albrecht import albrecht_7 as stroud_s2_15_2
from ._albrecht import albrecht_8 as stroud_s2_17_1
from ._albrecht_collatz import albrecht_collatz as stroud_s2_3_2
from ._hammer_stroud import hammer_stroud_11_2 as stroud_s2_3_1
from ._hammer_stroud import hammer_stroud_12_2 as stroud_s2_5_2
from ._hammer_stroud import hammer_stroud_18 as stroud_s2_7_2
from ._helpers import S2Scheme, register
from ._mysovskih import mysovskih_1 as stroud_s2_4_1
from ._mysovskih import mysovskih_2 as stroud_s2_11_1
from ._mysovskih import mysovskih_3 as stroud_s2_15_1
from ._peirce_1956 import peirce_1956_1 as stroud_s2_7_1
from ._peirce_1956 import peirce_1956_2 as stroud_s2_9_5
from ._peirce_1956 import peirce_1956_3 as stroud_s2_11_4
from ._rabinowitz_richter import rabinowitz_richter_1 as stroud_s2_9_2
from ._rabinowitz_richter import rabinowitz_richter_2 as stroud_s2_9_4
from ._rabinowitz_richter import rabinowitz_richter_4 as stroud_s2_11_3
from ._rabinowitz_richter import rabinowitz_richter_5 as stroud_s2_13_1
from ._radon import radon

_source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_s2_5_1():
    return radon(0)


def stroud_s2_9_3():
    # spherical product gauss 9
    sqrt = np.vectorize(sympy.sqrt)
    pm_ = np.array([+1, -1])
    cos = np.vectorize(sympy.cos)
    sin = np.vectorize(sympy.sin)
    frac = sympy.Rational
    pi = sympy.pi

    r1, r2 = sqrt((6 - pm_ * sqrt(6)) / 10)

    a = 2 * (np.arange(10) + 1) * pi / 10
    x = np.array([cos(a), sin(a)]).T

    B0 = frac(1, 9)
    B1, B2 = (16 + pm_ * sqrt(6)) / 360

    data = [(B0, z(2)), (B1, r1 * x), (B2, r2 * x)]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    d = {"plain": [weights, points[0], points[1]]}
    return S2Scheme("Stroud S2 9-3", d, 9, _source)


register(
    [
        stroud_s2_3_1,
        stroud_s2_3_2,
        stroud_s2_4_1,
        stroud_s2_5_1,
        stroud_s2_5_2,
        stroud_s2_7_1,
        stroud_s2_7_2,
        stroud_s2_9_1,
        stroud_s2_9_2,
        stroud_s2_9_3,
        stroud_s2_9_4,
        stroud_s2_9_5,
        stroud_s2_11_1,
        stroud_s2_11_2,
        stroud_s2_11_3,
        stroud_s2_11_4,
        stroud_s2_13_1,
        stroud_s2_13_2,
        stroud_s2_15_1,
        stroud_s2_15_2,
        stroud_s2_17_1,
    ]
)



from ._felippa import (
    felippa_1,
    felippa_2,
    felippa_3,
    felippa_4,
    felippa_5,
    felippa_6,
    felippa_7,
    felippa_8,
    felippa_9,
)

__all__ = [
    "felippa_1",
    "felippa_2",
    "felippa_3",
    "felippa_4",
    "felippa_5",
    "felippa_6",
    "felippa_7",
    "felippa_8",
    "felippa_9",
]



import pathlib

from sympy import Rational as frac
from sympy import sqrt

from ...helpers import article
from .._helpers import T3Scheme, _read, register

source = article(
    authors=["F.D. Witherden", "P.E. Vincent"],
    title="On the identification of symmetric quadrature rules for finite element methods",
    journal="Computers & Mathematics with Applications",
    volume="69",
    number="10",
    month="may",
    year="2015",
    pages="1232–1241",
    url="https://doi.org/10.1016/j.camwa.2015.03.017",
)

this_dir = pathlib.Path(__file__).resolve().parent


def witherden_vincent_01():
    degree = 1
    d = {"s4": [[1]]}
    return T3Scheme("Witherden-Vincent 1", d, degree, source)


def witherden_vincent_02():
    degree = 2
    d = {"s31": [[frac(1, 4)], [frac(1, 4) - sqrt(5) / 20]]}
    return T3Scheme("Witherden-Vincent 2", d, degree, source)


def witherden_vincent_03():
    return _read(this_dir / "witherden_vincent_03.json", source)


def witherden_vincent_05():
    return _read(this_dir / "witherden_vincent_05.json", source)


def witherden_vincent_06():
    return _read(this_dir / "witherden_vincent_06.json", source)


def witherden_vincent_07():
    return _read(this_dir / "witherden_vincent_07.json", source)


def witherden_vincent_08():
    return _read(this_dir / "witherden_vincent_08.json", source)


def witherden_vincent_09():
    return _read(this_dir / "witherden_vincent_09.json", source)


def witherden_vincent_10():
    return _read(this_dir / "witherden_vincent_10.json", source)


register(
    [
        witherden_vincent_01,
        witherden_vincent_02,
        witherden_vincent_03,
        witherden_vincent_05,
        witherden_vincent_06,
        witherden_vincent_07,
        witherden_vincent_08,
        witherden_vincent_09,
        witherden_vincent_10,
    ]
)



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import U3Scheme, register

source = article(
    authors=["J. Albrecht", "L. Collatz"],
    title="Zur numerischen Auswertung mehrdimensionaler Integrale",
    journal="ZAMM",
    volume="38",
    number="1-2",
    year="1958",
    pages="1–15",
    url="https://doi.org/10.1002/zamm.19580380102",
)


def albrecht_collatz_1():
    r, s = (sqrt((5 + t * sqrt(5)) / 10) for t in [+1, -1])
    data = {"rs0": [frac(1, 12), r, s]}
    return U3Scheme("Albrecht-Collatz 1", data, 5, source)


def albrecht_collatz_2():
    data = {"a1": [frac(8, 120)], "a3": [frac(9, 120)]}
    return U3Scheme("Albrecht-Collatz 2", data, 5, source)


def albrecht_collatz_3():
    data = {"a1": [frac(1, 30)], "a2": [frac(2, 30)]}
    return U3Scheme("Albrecht-Collatz 3", data, 5, source)


def albrecht_collatz_4():
    r, s = (sqrt((3 + t * sqrt(5)) / 6) for t in [+1, -1])
    data = {"rs0": [frac(1, 20), r, s], "a3": [frac(1, 20)]}
    return U3Scheme("Albrecht-Collatz 4", data, 5, source)


def albrecht_collatz_5():
    data = {
        "a1": [frac(40, 840)],
        "a2": [frac(32, 840)],
        "a3": [frac(27, 840)],
    }
    return U3Scheme("Albrecht-Collatz 5", data, 7, source)


register(
    [
        albrecht_collatz_1,
        albrecht_collatz_2,
        albrecht_collatz_3,
        albrecht_collatz_4,
        albrecht_collatz_5,
    ]
)



import pathlib

from sympy import Rational as frac

from ...helpers import article
from .._helpers import T2Scheme, _read, register

source = article(
    authors=["Mark A. Taylor", "Beth A. Wingate", "Len P. Bos"],
    title="Several new quadrature formulas for polynomial integration in the triangle",
    journal="arXiv Mathematics e-prints",
    year="2005",
    month="jan",
    url="https://arxiv.org/abs/math/0501496",
)

this_dir = pathlib.Path(__file__).resolve().parent


# TODO missing Taylor-Wingate-Bos schemes


def taylor_wingate_bos_1():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 6)]]}
    return T2Scheme("Taylor-Wingate-Bos 1", d, 2, source)


def taylor_wingate_bos_2():
    return _read(this_dir / "taylor_wingate_bos_2.json", source)


def taylor_wingate_bos_4():
    return _read(this_dir / "taylor_wingate_bos_4.json", source)


def taylor_wingate_bos_5():
    return _read(this_dir / "taylor_wingate_bos_5.json", source)


def taylor_wingate_bos_8():
    return _read(this_dir / "taylor_wingate_bos_8.json", source)


register(
    [
        taylor_wingate_bos_1,
        taylor_wingate_bos_2,
        taylor_wingate_bos_4,
        taylor_wingate_bos_5,
        taylor_wingate_bos_8,
    ]
)



from . import (
    c1,
    c2,
    c3,
    cn,
    e1r,
    e1r2,
    e2r,
    e2r2,
    e3r,
    e3r2,
    enr,
    enr2,
    helpers,
    p3,
    s2,
    s3,
    sn,
    t2,
    t3,
    tn,
    tools,
    u2,
    u3,
    w3,
)
from .__about__ import __version__
from ._optimize import optimize
from ._scipy_compat import quad

__all__ = [
    "__version__",
    "s3",
    "u2",
    "s2",
    "e1r",
    "e1r2",
    "e2r",
    "e2r2",
    "e3r",
    "e3r2",
    "enr",
    "enr2",
    "helpers",
    "c3",
    "c1",
    "sn",
    "cn",
    "tn",
    "p3",
    "c2",
    "u3",
    "t3",
    "tools",
    "t2",
    "w3",
    "tools",
    "quad",
    "optimize",
]



from ..tn import get_vol, transform
from . import (
    _albrecht_collatz,
    _berntsen_espelid,
    _classical,
    _cools_haegemans,
    _cubtri,
    _dunavant,
    _franke,
    _gatermann,
    _griener_schmid,
    _hammer_marlowe_stroud,
    _hammer_stroud,
    _hillion,
    _laursen_gellert,
    _lether,
    _liu_vinokur,
    _lyness_jespersen,
    _papanicolopulos,
    _strang_fix_cowper,
    _stroud,
    _taylor_wingate_bos,
    _vioreanu_rokhlin,
    _walkington,
    _wandzura_xiao,
    _williams_shunn_jameson,
    _witherden_vincent,
    _xiao_gimbutas,
    _zhang_cui_liu,
)
from ._helpers import get_good_scheme, schemes
from ._tools import integrate_adaptive

__all__ = [
    "integrate_adaptive",
    "transform",
    "get_vol",
    "schemes",
    "get_good_scheme",
    #
    "_albrecht_collatz",
    "_berntsen_espelid",
    "_classical",
    "_cools_haegemans",
    "_cubtri",
    "_dunavant",
    "_franke",
    "_gatermann",
    "_griener_schmid",
    "_hammer_marlowe_stroud",
    "_hammer_stroud",
    "_hillion",
    "_laursen_gellert",
    "_lether",
    "_liu_vinokur",
    "_lyness_jespersen",
    "_papanicolopulos",
    "_strang_fix_cowper",
    "_stroud",
    "_taylor_wingate_bos",
    "_vioreanu_rokhlin",
    "_walkington",
    "_wandzura_xiao",
    "_williams_shunn_jameson",
    "_witherden_vincent",
    "_xiao_gimbutas",
    "_zhang_cui_liu",
]



from sympy import Rational as frac

from ..helpers import book
from ._helpers import C2Scheme, register

source = book(
    authors=["Joseph Oscar Irwin"],
    title="On quadrature and cubature",
    publisher="Cambridge University Press",
    year="1923",
    url="https://books.google.de/books/about/On_quadrature_and_cubature.html?id=SuruAAAAMAAJ",
)


def irwin_1():
    d = {"d4_aa": [[frac(14, 48)], [1]], "d4_ab": [[-frac(1, 48)], [3], [1]]}
    return C2Scheme("Irwin 1", d, 3, source)


def irwin_2():
    d = {
        "d4_aa": [[frac(889, 2880), frac(5, 2880)], [1, 3]],
        "d4_ab": [[-frac(98, 2880), frac(11, 2880)], [3, 5], [1, 1]],
    }
    return C2Scheme("Irwin 2", d, 5, source, 5.685e-14)


register([irwin_1, irwin_2])



import numpy as np
from mpmath import mp

from ..helpers import article, rd, untangle
from ._helpers import TnScheme

source = article(
    authors=["A.H. Stroud"],
    title="Approximate Integration Formulas of Degree 3 for Simplexes",
    journal="Mathematics of Computation",
    volume="18",
    number="88",
    month="oct",
    year="1964",
    pages="590-597",
    url="https://doi.org/10.2307/2002945",
)


def _stroud_1964(variant_a, n):
    degree = 3

    if n == 2:
        # The roots sum up to 1
        r, s, t = mp.polyroots([1, -1, 0.25, -1.0 / 60.0])
        data = [(1.0 / (n * (n + 1)), rd(n + 1, [(r, 1), (s, 1), (t, 1)]))]
    else:
        assert n > 2

        # Stroud's book only gives numerical values for certain n; the article explains
        # it in more detail, namely: r is a root of a polynomial of degree 3.
        rts = np.sort(np.roots([n + 1, -3, 3 / (n + 2), -1 / ((n + 2) * (n + 3))]))

        # all roots are real-valued
        if n > 8:
            assert not variant_a, "Choose variant b for n >= 9."

        r = rts[0] if variant_a else rts[1]

        # s and t are zeros of a polynomial of degree 2
        s, t = np.sort(
            np.roots(
                [
                    1,
                    -(1 - (n - 1) * r),
                    n / (2 * (n + 2)) - (n - 1) * r + n * (n - 1) / 2 * r ** 2,
                ]
            )
        )

        data = [(1 / (n * (n + 1)), rd(n + 1, [(r, n - 1), (s, 1), (t, 1)]))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)

    variant = "a" if variant_a else "b"
    name = f"Stroud 1964{variant}"
    return TnScheme(name, n, weights, points, degree, source)


def stroud_1964a(n):
    return _stroud_1964(True, n)


def stroud_1964b(n):
    return _stroud_1964(False, n)



import json

import numpy as np
import sympy

from ..helpers import backend_to_function
from ..tn import TnScheme, get_vol, transform

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class T3Scheme(TnScheme):
    def __init__(
        self, name, symmetry_data, degree, source=None, tol=1.0e-14, weight_factor=None
    ):
        self.symmetry_data = symmetry_data
        points, weights = expand_symmetries(symmetry_data)
        if weight_factor is not None:
            weights *= weight_factor

        super().__init__(name, 3, weights, points, degree, source, tol)
        self.domain = "T3"

    def show(
        self,
        tet=np.array(
            [
                [+1, 0, -1.0 / np.sqrt(2.0)],
                [-1, 0, -1.0 / np.sqrt(2.0)],
                [0, +1, +1.0 / np.sqrt(2.0)],
                [0, -1, +1.0 / np.sqrt(2.0)],
            ]
        ),
        backend="vtk",
        render=True,
    ):
        edges = np.array([[tet[i], tet[j]] for i in range(4) for j in range(i)])
        edges = np.moveaxis(edges, 1, 2)
        backend_to_function[backend](
            transform(self.points.T, tet.T).T,
            self.weights,
            get_vol(tet),
            edges,
            render=render,
        )


def _s4(dummy):
    if dummy.dtype == sympy.Basic:
        return np.full((4, 1), sympy.Rational(1, 4))
    return np.full((4, 1), 0.25)


def _s31(a):
    b = 1 - 3 * a
    points = np.array([[a, a, a, b], [a, a, b, a], [a, b, a, a], [b, a, a, a]])
    points = np.moveaxis(points, 0, 1)
    return points


def _s22(a):
    if a.dtype in [sympy.Basic, int]:
        b = (1 - 2 * a) / sympy.S(2)
    else:
        b = (1 - 2 * a) / 2
    points = np.array(
        [
            [a, a, b, b],
            [a, b, a, b],
            [b, a, a, b],
            [a, b, b, a],
            [b, a, b, a],
            [b, b, a, a],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _s211(data):
    a, b = data
    c = 1 - 2 * a - b
    points = np.array(
        [
            [a, a, b, c],
            [a, b, a, c],
            [b, a, a, c],
            [a, b, c, a],
            [b, a, c, a],
            [b, c, a, a],
            [a, a, c, b],
            [a, c, a, b],
            [c, a, a, b],
            [a, c, b, a],
            [c, a, b, a],
            [c, b, a, a],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _s1111(data):
    a, b, c = data
    d = 1 - a - b - c
    points = np.array(
        [
            [a, b, c, d],
            [a, b, d, c],
            [a, c, b, d],
            [a, c, d, b],
            [a, d, b, c],
            [a, d, c, b],
            [b, a, c, d],
            [b, a, d, c],
            [b, c, a, d],
            [b, c, d, a],
            [b, d, a, c],
            [b, d, c, a],
            [c, a, b, d],
            [c, a, d, b],
            [c, b, a, d],
            [c, b, d, a],
            [c, d, a, b],
            [c, d, b, a],
            [d, a, b, c],
            [d, a, c, b],
            [d, b, a, c],
            [d, b, c, a],
            [d, c, a, b],
            [d, c, b, a],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def expand_symmetries_points_only(data):
    points = []
    counts = []

    for key, points_raw in data.items():
        fun = {
            "s4": _s4,
            "s31": _s31,
            "s211": _s211,
            "s22": _s22,
            "s1111": _s1111,
            "plain": lambda vals: vals.reshape(4, 1, -1),
        }[key]
        pts = fun(np.asarray(points_raw))

        counts.append(pts.shape[1])
        pts = pts.reshape(pts.shape[0], -1)
        points.append(pts)

    points = np.ascontiguousarray(np.concatenate(points, axis=1))
    return points, counts


def expand_symmetries(data):
    # separate points and weights
    points_raw = {}
    weights_raw = []
    for key, values in data.items():
        values = np.asarray(values)
        weights_raw.append(values[0])
        points_raw[key] = values[1:]

    points, counts = expand_symmetries_points_only(points_raw)
    weights = np.concatenate(
        [np.tile(values, count) for count, values in zip(counts, weights_raw)]
    )
    return points, weights


def _read(filepath, source):
    with open(filepath) as f:
        content = json.load(f)

    degree = content["degree"]
    name = content["name"]
    tol = content["test_tolerance"]
    data = content["data"]
    weight_factor = content["weight factor"] if "weight factor" in content else None
    return T3Scheme(name, data, degree, source, tol, weight_factor=weight_factor)



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..enr2._stroud_secrest import _nsimplex
from ..helpers import article, fsd, pm, untangle
from ._helpers import EnrScheme

source = article(
    authors=["A.H. Stroud", "D. Secrest"],
    title="Approximate integration formulas for certain spherically symmetric regions",
    journal="Math. Comp.",
    volume="17",
    year="1963",
    pages="105-135",
    url="https://doi.org/10.1090/S0025-5718-1963-0161473-0",
)


def stroud_secrest_1(n):
    data = [(frac(1, n + 1), sqrt(n + 1) * _nsimplex(n))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return EnrScheme("Stroud-Secrest I", n, weights, points, 2, source, 4.676e-14)


def stroud_secrest_2(n):
    nu = sqrt(n * (n + 1))
    data = [(frac(1, 2 * n), fsd(n, (nu, 1)))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return EnrScheme("Stroud-Secrest II", n, weights, points, 3, source)


def stroud_secrest_3(n):
    nu = sqrt(n + 1)
    data = [(frac(1, 2 ** n), pm(n * [nu]))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return EnrScheme("Stroud-Secrest III", n, weights, points, 3, source)


def stroud_secrest_4(n):
    nu = sqrt((n + 2) * (n + 3))
    xi = sqrt(frac((n + 2) * (n + 3), 2))
    A = frac(2 * (2 * n + 3), (n + 2) * (n + 3))
    B = frac((4 - n) * (n + 1), 2 * (n + 2) ** 2 * (n + 3))
    C = frac(n + 1, (n + 2) ** 2 * (n + 3))

    data = [(A, np.full((1, n), 0)), (B, fsd(n, (nu, 1))), (C, fsd(n, (xi, 2)))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return EnrScheme("Stroud-Secrest IV", n, weights, points, 5, source)



import math

import numpy as np
import sympy

from ..helpers import article, expand_symmetries, prod
from ._helpers import CnScheme

_source = article(
    authors=["Ronald Cools", "Ann Haegemans"],
    title="An imbedded family of cubature formulae for n-dimensional product regions",
    journal="Journal of Computational and Applied Mathematics",
    volume="51",
    year="1994",
    pages="251-262",
    url="https://doi.org/10.1016/0377-0427%2892%2900007-V",
)


def _mu(j, symbolic):
    # 1/2 int_{-1}^1 x^j
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    return frac(1, j + 1) if j % 2 == 0 else 0


def cools_haegemans_1(n, delta2=1, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    assert frac(1, 3) <= delta2
    m = 1

    w0 = frac(3 * delta2 - 1, 3 * delta2)
    w = frac(1, 3) ** m / (2 ** n * delta2 ** m)

    d = {"0": [[w0]], "a": [[w], [sqrt(delta2)]]}
    points, weights = expand_symmetries(d, n)
    return CnScheme(f"Cools-Haegemans 1 (dim={n})", n, weights, points, 3, _source)


def cools_haegemans_2(n, delta2=1, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    # assert frac(5, 11) <= delta2 <= frac(5 * n + 4, 3 * (5 * n - 4))
    m = 2

    lmbdas2 = _gener(delta2, 1, _mu, symbolic)
    w0 = frac(
        -(15 * delta2 * n - 12 * delta2 - 5 * n - 4) * (3 * delta2 - 1),
        36 * delta2 ** 2,
    )
    w1 = frac(5 * (3 * delta2 - 1) ** 2, 72 * delta2)
    w = frac(1, 3) ** m / (2 ** n * delta2 ** m)

    lmbdas = [sqrt(lmbda2) for lmbda2 in lmbdas2]

    d = {"0": [[w0]], "a0": [[w1], [lmbdas[0]]], "a": [[w], [sqrt(delta2)]]}
    points, weights = expand_symmetries(d, n)
    return CnScheme("Cools-Haegemans 2", n, weights, points, 5, _source, 6.312e-14)


# ERR There is a mistake here somewhere in the weights, but it's unclear where.
# TODO fix this
# def cools_haegemans_3(n, delta2=1):
#     assert n >= 2
#     m = 3
#
#     lmbdas2 = _gener(delta2, 2, _mu)
#
#     delta4 = delta2 ** 2
#     delta6 = delta2 ** 3
#     w1 = frac(
#         -5
#         * (3 * delta2 - 1)
#         * (
#             135 * delta4 * n
#             - 68 * delta4
#             - 90 * delta2 * n
#             - 120 * delta2
#             + 15 * n
#             + 60
#         ),
#         2592 * delta6,
#     )
#     w2 = frac(
#         245 * (5 * delta2 - 3) ** 3 * (3 * delta2 - 1),
#         5184 * (31 * delta2 - 15) * delta6,
#     )
#     w11 = frac(25 * (3 * delta2 - 1) ** 3, 1728 * delta6)
#     w0 = frac(
#         -27 * (2 * n * (w1 + w2 - w11) + 2 * n ** 2 * w11 - 1) * delta6 + 1,
#         27 * delta6,
#     )
#     w = frac(_mu(2) ** m * _mu(0) ** (n - m), 2 ** n * delta2 ** m)
#
#     delta = sqrt(delta2)
#     lmbdas = [sqrt(lmbda2) for lmbda2 in lmbdas2]
#
#     # It's not entirely clear which points are meant to be part of the scheme. The
#     # weights add up to more than 1.
#     data = [
#         (w0, z(n)),
#         (w1, fsd(n, (lmbdas[0], 1))),
#         (w2, fsd(n, (lmbdas[1], 1))),
#         (w11, fsd(n, (lmbdas[0], 2))),
#         (w, pm(n, delta)),
#     ]
#
#     # This is an attempt to find the correct symmetries, but to no avail. Something
#     # appears to be wrong with the weights.
#     # print(n)
#     # maxk = 20
#     # k0 = 1
#     # k = 2 ** n
#     # print("try")
#     # for k1 in range(0, maxk, 2):
#     #     for k2 in range(0, maxk, 2):
#     #         for k11 in range(0, maxk, 2):
#     #             val = k0 * w0 + k1 * w1 + k2 * w2 + k11 * w11 + k * w
#     #             if val == 1:
#     #                 print("SUCCESS", k0, k1, k2, k11, k)
#     # exit(1)
#
#     points, weights = untangle(data)
#
#     print(points)
#     print(weights)
#     print(sum(weights))
#     exit(1)
#     return CnScheme("Cools-Haegemans 3", n, weights, points, 7, _source)


def _gener(delta2, m, mu, symbolic):
    # Computes the lambda_q from the article, eq. (9).
    lmbdas2 = []
    for q in range(1, m + 1):
        if not lmbdas2:
            # https://github.com/numpy/numpy/issues/16152
            coeffs = [1]
        else:
            coeffs = np.poly(lmbdas2)

        a0 = [c * mu(2 * (q - k) + 2, symbolic) for k, c in enumerate(coeffs)]
        a1 = [c * mu(2 * (q - k), symbolic) for k, c in enumerate(coeffs)]
        prod_ = prod([1 - lmbda2 / delta2 for lmbda2 in lmbdas2])
        a = sum(a0) - mu(2, symbolic) ** (q + 1) / mu(0, symbolic) ** q * prod_
        b = sum(a1) - mu(2, symbolic) ** (q + 1) / mu(0, symbolic) ** q * prod_ / delta2
        lmbdas2.append(a / b)
    return lmbdas2



from ..cn import ncube_points as rectangle_points
from ..cn import transform
from . import (
    _albrecht_collatz,
    _burnside,
    _cohen_gismalla,
    _cools_haegemans_1985,
    _cools_haegemans_1988,
    _dunavant,
    _franke,
    _haegemans_piessens,
    _hammer_stroud,
    _irwin,
    _maxwell,
    _meister,
    _miller,
    _morrow_patterson,
    _phillips,
    _piessens_haegemans,
    _rabinowitz_richter,
    _schmid,
    _sommariva,
    _stroud,
    _tyler,
    _waldron,
    _wissmann_becker,
    _witherden_vincent,
)
from ._helpers import get_good_scheme, schemes
from ._product import product

__all__ = [
    "_albrecht_collatz",
    "_burnside",
    "_cohen_gismalla",
    "_cools_haegemans_1985",
    "_cools_haegemans_1988",
    "_dunavant",
    "_franke",
    "_haegemans_piessens",
    "_hammer_stroud",
    "_irwin",
    "_maxwell",
    "_meister",
    "_miller",
    "_morrow_patterson",
    "_phillips",
    "_piessens_haegemans",
    "_rabinowitz_richter",
    "_schmid",
    "_sommariva",
    "_stroud",
    "_tyler",
    "_waldron",
    "_wissmann_becker",
    "_witherden_vincent",
    "product",
    #
    "transform",
    "rectangle_points",
    "schemes",
    "get_good_scheme",
]



from math import factorial as fact

import ndim
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, fsd, untangle
from ._helpers import UnScheme

source = article(
    authors=["L.N. Dobrodeev"],
    title="Cubature rules with equal coefficients for integrating functions with respect to symmetric domains",
    journal="USSR Computational Mathematics and Mathematical Physics",
    volume="18",
    number="4",
    year="1978",
    pages="27-34",
    url="https://doi.org/10.1016/0041-5553%2878%2990064-2",
)


def dobrodeev_1978(n: int):
    # from sympy import sqrt, factorial as fact, Rational as fr

    assert 2 <= n <= 20

    degree = 5

    dim_config = {
        2: ("I", None, 1, 1),
        3: ("I", None, 1, 1),
        4: (None, 2, None, None),
        5: ("I", 5, 1, 1),
        6: ("I", 5, 1, 1),
        7: (None, 3, None, None),
        8: ("I", 8, 1, 1),
        9: ("I", 9, 1, 1),
        10: ("I", 10, 1, 1),
        11: ("I", 11, 1, 1),
        12: ("I", 5, 1, 1),
        13: ("I", 13, 1, 2),
        14: ("I", 14, 1, 2),
        15: ("I", 15, 1, 2),
        16: ("I", 16, 1, 2),
        17: ("I", 17, 1, 2),
        18: ("I", 18, 1, 3),
        19: ("I", 19, 1, 3),
        20: ("I", 20, 1, 3),
    }

    pm_type, i, j, k = dim_config[n]

    I0 = ndim.nsphere.integrate_monomial(n * [0], symbolic=True)

    if i is None:
        G, b, c = _generate_jk(n, pm_type, j, k)
        data = [(G, fsd(n, (b, j), (c, k)))]
    elif j is None:
        assert k is None
        assert pm_type is None
        G, a = _generate_i(n, i)
        data = [(G, fsd(n, (a, i)))]
    else:
        I2 = ndim.nsphere.integrate_monomial([2] + (n - 1) * [0], symbolic=True)
        I22 = ndim.nsphere.integrate_monomial([2, 2] + (n - 2) * [0], symbolic=True)
        I4 = ndim.nsphere.integrate_monomial([4] + (n - 1) * [0], symbolic=True)
        G, a, b, c = _compute_dobrodeev(n, I0, I2, I22, I4, pm_type, i, j, k)

        data = [(G, fsd(n, (a, i))), (G, fsd(n, (b, j), (c, k)))]

    points, weights = untangle(data)
    return UnScheme("Dobrodeev 1978", n, weights, points, degree, source)


def _generate_i(n: int, i: int):
    L = fact(n) // fact(i) // fact(n - i) * 2 ** i
    G = frac(1, L)
    a = sqrt(frac(3, n + 2))
    return G, a


def _generate_jk(n, pm_type, j, k):
    M = fact(n) // fact(j) // fact(k) // fact(n - j - k) * 2 ** (j + k)
    G = frac(1, M)

    t = 1 if pm_type == "I" else -1
    b = sqrt(frac(1, j + k) * (1 + t * (k / j * sqrt(3 * (j + k) / (n + 2) - 1))))
    c = sqrt(frac(1, j + k) * (1 - t * (j / k * sqrt(3 * (j + k) / (n + 2) - 1))))
    return G, b, c


def _compute_dobrodeev(n, I0, I2, I22, I4, pm_type, i, j, k):
    """Same as the helper function in ..helpers, making use of the fact that
    `F == 0` for the sphere
    """
    # TODO prove F==0 analytically
    t = 1 if pm_type == "I" else -1

    L = fact(n) // (fact(i) * fact(n - i)) * 2 ** i
    M = fact(n) // (fact(j) * fact(k) * fact(n - j - k)) * 2 ** (j + k)
    N = L + M
    R = (
        -frac(j + k - i, i) * I2 ** 2 / I0 ** 2
        + frac(j + k - 1, n) * I4 / I0
        - frac(n - 1, n) * I22 / I0
    )
    H = frac(1, i) * (
        +(j + k - i) * I2 ** 2 / I0 ** 2
        + frac(j + k, n) * ((i - 1) * I4 / I0 - (n - 1) * I22 / I0)
    )
    Q = L / M * R + H

    G = frac(1, N)
    a = sqrt(frac(n, i) * I2 / I0)
    b = sqrt(frac(n, j + k) * (I2 / I0 + t * sqrt(k / j * Q)))
    c = sqrt(frac(n, j + k) * (I2 / I0 - t * sqrt(j / k * Q)))
    return G, a, b, c



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["D.M. Williams", "L. Shunn", "A. Jameson"],
    title="Symmetric quadrature rules for simplexes based on sphere close packed lattice arrangements",
    journal="Journal of Computational and Applied Mathematics",
    volume="266",
    year="2014",
    pages="18–38",
    url="https://doi.org/10.1016/j.cam.2014.01.007",
)

this_dir = pathlib.Path(__file__).resolve().parent


def williams_shunn_jameson():
    return _read(this_dir / "williams_shunn_jameson.json", source)


register([williams_shunn_jameson])



# TODO all points are "plain" -- identify symmetries and list them as such
import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["Hong Xiao", "Zydrunas Gimbutas"],
    title="A numerical algorithm for the construction of efficient quadrature rules in two and higher dimensions",
    journal="Computers & Mathematics with Applications",
    volume="59",
    number="2",
    month="jan",
    year="2010",
    pages="663–676",
    url="https://doi.org/10.1016/j.camwa.2009.10.027",
)
# Data extracted from
# https://people.sc.fsu.edu/~jburkardt/f_src/triangle_symq_rule/triangle_symq_rule.f90

this_dir = pathlib.Path(__file__).resolve().parent


def xiao_gimbutas_01():
    return _read(this_dir / "xg01.json", source)


def xiao_gimbutas_02():
    return _read(this_dir / "xg02.json", source)


def xiao_gimbutas_03():
    return _read(this_dir / "xg03.json", source)


def xiao_gimbutas_04():
    return _read(this_dir / "xg04.json", source)


def xiao_gimbutas_05():
    return _read(this_dir / "xg05.json", source)


def xiao_gimbutas_06():
    return _read(this_dir / "xg06.json", source)


def xiao_gimbutas_07():
    return _read(this_dir / "xg07.json", source)


def xiao_gimbutas_08():
    return _read(this_dir / "xg08.json", source)


def xiao_gimbutas_09():
    return _read(this_dir / "xg09.json", source)


def xiao_gimbutas_10():
    return _read(this_dir / "xg10.json", source)


def xiao_gimbutas_11():
    return _read(this_dir / "xg11.json", source)


def xiao_gimbutas_12():
    return _read(this_dir / "xg12.json", source)


def xiao_gimbutas_13():
    return _read(this_dir / "xg13.json", source)


def xiao_gimbutas_14():
    return _read(this_dir / "xg14.json", source)


def xiao_gimbutas_15():
    return _read(this_dir / "xg15.json", source)


register(
    [
        xiao_gimbutas_01,
        xiao_gimbutas_02,
        xiao_gimbutas_03,
        xiao_gimbutas_04,
        xiao_gimbutas_05,
        xiao_gimbutas_06,
        xiao_gimbutas_07,
        xiao_gimbutas_08,
        xiao_gimbutas_09,
        xiao_gimbutas_10,
        xiao_gimbutas_11,
        xiao_gimbutas_12,
        xiao_gimbutas_13,
        xiao_gimbutas_14,
        xiao_gimbutas_15,
    ]
)



import numpy as np

from ._helpers import C1Scheme


def trapezoidal():
    weights = np.array([1.0, 1.0])
    points = np.array([-1.0, 1.0])
    return C1Scheme("Trapezoidal rule", 1, weights, points)



import pathlib

from ...helpers import article
from .._helpers import _read, register

_source = article(
    authors=["Ann Haegemans", "Robert Piessens"],
    title="Construction of Cubature Formulas of Degree Seven and Nine Symmetric Planar Regions Using Orthogonal Polynomials",
    journal="SIAM Journal on Numerical Analysis",
    volume="14",
    number="3",
    month="jun",
    year="1977",
    pages="492-508",
    url="https://www.jstor.org/stable/2156699",
)

this_dir = pathlib.Path(__file__).resolve().parent


def haegemans_piessens():
    return _read(this_dir / "haegemans_piessens.json", _source)


register([haegemans_piessens])



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ..un._mysovskikh import get_nsimplex_points
from ._helpers import Enr2Scheme
from ._phillips import phillips as lu_darmofal_3
from ._stroud import stroud_enr2_5_1a as lu_darmofal_4a
from ._stroud import stroud_enr2_5_1b as lu_darmofal_4b
from ._stroud_secrest import stroud_secrest_4 as lu_darmofal_2

source = article(
    authors=["James Lu", "David L. Darmofal"],
    title="Higher-Dimensional Integration with Gaussian Weight for Applications in Probabilistic Design",
    journal="SIAM J. Sci. Comput.",
    volume="26",
    number="2",
    year="2004",
    pages="613–624",
    url="https://doi.org/10.1137/S1064827503426863",
)


def lu_darmofal_1(n):
    # ENH The article says n>=4, but the scheme also works for 2, 3
    assert n >= 2
    a = get_nsimplex_points(n, sqrt, frac)
    b = np.array(
        [
            sqrt(frac(n, 2 * (n - 1))) * (a[k] + a[l])
            for k in range(len(a))
            for l in range(k)
        ]
    )
    points = np.concatenate(
        [
            [[0] * n],
            +sqrt(frac(n, 2) + 1) * a,
            -sqrt(frac(n, 2) + 1) * a,
            +sqrt(frac(n, 2) + 1) * b,
            -sqrt(frac(n, 2) + 1) * b,
        ]
    )
    points = np.ascontiguousarray(points.T)

    p = frac(2, n + 2)
    A = frac(n ** 2 * (7 - n), 2 * (n + 1) ** 2 * (n + 2) ** 2)
    B = frac(2 * (n - 1) ** 2, (n + 1) ** 2 * (n + 2) ** 2)
    weights = np.concatenate(
        [
            [p],
            np.full(len(a), A),
            np.full(len(a), A),
            np.full(len(b), B),
            np.full(len(b), B),
        ]
    )
    return Enr2Scheme("Lu-Darmofal I", n, weights, points, 5, source)


__all__ = [
    "lu_darmofal_1",
    "lu_darmofal_2",
    "lu_darmofal_3",
    "lu_darmofal_4a",
    "lu_darmofal_4b",
]



from sympy import Rational as frac

from ..helpers import article
from ._helpers import C3Scheme, register

_source = article(
    authors=["D. Mustard", "J.N. Lyness", "J.M. Blatt"],
    title="Numerical quadrature in n dimensions",
    journal="The Computer Journal",
    year="1963",
    volume="6",
    number="1",
    pages="75-87",
    url="https://doi.org/10.1093/comjnl/6.1.75",
)


def mustard_lyness_blatt_1():
    d = {"zero3": [[frac(1, 2)]], "symm_rr0": [[frac(1, 24)], [1]]}
    return C3Scheme("Mustard-Lyness-Blatt 1", d, 3, _source)


def mustard_lyness_blatt_2():
    d = {
        "zero3": [[frac(2, 9)]],
        "symm_r00": [[frac(1, 9)], [1]],
        "symm_rrr": [[frac(1, 72)], [1]],
    }
    return C3Scheme("Mustard-Lyness-Blatt 2", d, 3, _source)


def mustard_lyness_blatt_3():
    d = {
        "symm_rr0": [[frac(1, 6)], [1]],
        "symm_rrr": [[-frac(1, 8)], [1]],
    }
    return C3Scheme("Mustard-Lyness-Blatt 3", d, 3, _source)


def mustard_lyness_blatt_4():
    d = {
        "zero3": [[-frac(2, 45)]],
        "symm_r00": [[+frac(2, 45)], [1]],
        "symm_rrr": [[frac(4, 45), frac(1, 120)], [frac(1, 2), 1]],
    }
    return C3Scheme("Mustard-Lyness-Blatt 4", d, 5, _source)


def mustard_lyness_blatt_5():
    d = {
        "zero3": [[-frac(19, 15)]],
        "symm_r00": [[frac(16, 45), -frac(1, 30)], [frac(1, 2), 1]],
        "symm_rr0": [[frac(1, 36)], [1]],
    }
    return C3Scheme("Mustard-Lyness-Blatt 5", d, 5, _source)


def mustard_lyness_blatt_6():
    d = {
        "zero3": [[-frac(4, 3)]],
        "symm_r00": [[frac(16, 45)], [frac(1, 2)]],
        "symm_rr0": [[frac(1, 90)], [1]],
        "symm_rrr": [[frac(1, 120)], [1]],
    }
    return C3Scheme("Mustard-Lyness-Blatt 6", d, 5, _source)


def mustard_lyness_blatt_7():
    d = {
        "zero3": [[frac(2, 45)]],
        "symm_rr0": [[frac(1, 45)], [1]],
        "symm_rrr": [[frac(4, 45), frac(-1, 360)], [frac(1, 2), 1]],
    }
    return C3Scheme("Mustard-Lyness-Blatt 7", d, 5, _source)


register(
    [
        mustard_lyness_blatt_1,
        mustard_lyness_blatt_2,
        mustard_lyness_blatt_3,
        mustard_lyness_blatt_4,
        mustard_lyness_blatt_5,
        mustard_lyness_blatt_6,
        mustard_lyness_blatt_7,
    ]
)



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, fsd, pm, untangle
from ._helpers import Enr2Scheme

source = article(
    authors=["A.H. Stroud", "D. Secrest"],
    title="Approximate integration formulas for certain spherically symmetric regions",
    journal="Math. Comp.",
    volume="17",
    year="1963",
    pages="105-135",
    url="https://doi.org/10.1090/S0025-5718-1963-0161473-0",
)


def stroud_secrest_1(n):
    # TODO check which is more appropriate
    # print(_nsimplex(n))
    # print()
    # print(get_nsimplex_points(n))
    data = [(frac(1, n + 1), sqrt(frac(1, 2)) * _nsimplex(n))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Stroud-Secrest I", n, weights, points, 2, source)


def stroud_secrest_2(n):
    nu = sqrt(frac(n, 2))
    data = [(frac(1, 2 * n), fsd(n, (nu, 1)))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Stroud-Secrest II", n, weights, points, 3, source)


def stroud_secrest_3(n):
    nu = sqrt(frac(1, 2))
    data = [(frac(1, 2 ** n), pm(n * [nu]))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Stroud-Secrest III", n, weights, points, 3, source)


def stroud_secrest_4(n):
    nu = sqrt(frac(n + 2, 2))
    xi = sqrt(frac(n + 2, 4))
    A = frac(2, n + 2)
    B = frac(4 - n, 2 * (n + 2) ** 2)
    C = frac(1, (n + 2) ** 2)

    data = [(A, np.full((1, n), 0)), (B, fsd(n, (nu, 1))), (C, fsd(n, (xi, 2)))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Stroud-Secrest IV", n, weights, points, 5, source)


def _nsimplex(n):
    # construct the regular n-simplex points with 0 center
    return np.array(
        [
            [-sqrt(frac(n + 1, (n + 1 - k) * (n - k))) for k in range(i)]
            + [sqrt(frac((n + 1) * (n - i), n + 1 - i))]
            + (n - i - 1) * [0]
            for i in range(n)
        ]
        + [[-sqrt(frac(n + 1, (n + 1 - i) * (n - i))) for i in range(n)]]
    )



import pathlib

from ...helpers import online
from .._helpers import _read, register

_source = online(
    authors=["Alvise Sommariva"],
    year="2012",
    url="https://www.math.unipd.it/~alvise/sets.html",
)

this_dir = pathlib.Path(__file__).resolve().parent


def sommariva_01():
    return _read(this_dir / "sommariva_01.json", _source)


def sommariva_02():
    return _read(this_dir / "sommariva_02.json", _source)


def sommariva_03():
    return _read(this_dir / "sommariva_03.json", _source)


def sommariva_04():
    return _read(this_dir / "sommariva_04.json", _source)


def sommariva_05():
    return _read(this_dir / "sommariva_05.json", _source)


def sommariva_06():
    return _read(this_dir / "sommariva_06.json", _source)


def sommariva_07():
    return _read(this_dir / "sommariva_07.json", _source)


def sommariva_08():
    return _read(this_dir / "sommariva_08.json", _source)


def sommariva_09():
    return _read(this_dir / "sommariva_09.json", _source)


def sommariva_10():
    return _read(this_dir / "sommariva_10.json", _source)


def sommariva_11():
    return _read(this_dir / "sommariva_11.json", _source)


def sommariva_12():
    return _read(this_dir / "sommariva_12.json", _source)


def sommariva_13():
    return _read(this_dir / "sommariva_13.json", _source)


def sommariva_14():
    return _read(this_dir / "sommariva_14.json", _source)


def sommariva_15():
    return _read(this_dir / "sommariva_15.json", _source)


def sommariva_16():
    return _read(this_dir / "sommariva_16.json", _source)


def sommariva_17():
    return _read(this_dir / "sommariva_17.json", _source)


def sommariva_18():
    return _read(this_dir / "sommariva_18.json", _source)


def sommariva_19():
    return _read(this_dir / "sommariva_19.json", _source)


def sommariva_20():
    return _read(this_dir / "sommariva_20.json", _source)


def sommariva_21():
    return _read(this_dir / "sommariva_21.json", _source)


def sommariva_22():
    return _read(this_dir / "sommariva_22.json", _source)


def sommariva_23():
    return _read(this_dir / "sommariva_23.json", _source)


def sommariva_24():
    return _read(this_dir / "sommariva_24.json", _source)


def sommariva_25():
    return _read(this_dir / "sommariva_25.json", _source)


def sommariva_26():
    return _read(this_dir / "sommariva_26.json", _source)


def sommariva_27():
    return _read(this_dir / "sommariva_27.json", _source)


def sommariva_28():
    return _read(this_dir / "sommariva_28.json", _source)


def sommariva_29():
    return _read(this_dir / "sommariva_29.json", _source)


def sommariva_30():
    return _read(this_dir / "sommariva_30.json", _source)


def sommariva_31():
    return _read(this_dir / "sommariva_31.json", _source)


def sommariva_32():
    return _read(this_dir / "sommariva_32.json", _source)


def sommariva_33():
    return _read(this_dir / "sommariva_33.json", _source)


def sommariva_34():
    return _read(this_dir / "sommariva_34.json", _source)


def sommariva_35():
    return _read(this_dir / "sommariva_35.json", _source)


def sommariva_36():
    return _read(this_dir / "sommariva_36.json", _source)


def sommariva_37():
    return _read(this_dir / "sommariva_37.json", _source)


def sommariva_38():
    return _read(this_dir / "sommariva_38.json", _source)


def sommariva_39():
    return _read(this_dir / "sommariva_39.json", _source)


def sommariva_40():
    return _read(this_dir / "sommariva_40.json", _source)


def sommariva_41():
    return _read(this_dir / "sommariva_41.json", _source)


def sommariva_42():
    return _read(this_dir / "sommariva_42.json", _source)


def sommariva_43():
    return _read(this_dir / "sommariva_43.json", _source)


def sommariva_44():
    return _read(this_dir / "sommariva_44.json", _source)


def sommariva_45():
    return _read(this_dir / "sommariva_45.json", _source)


def sommariva_46():
    return _read(this_dir / "sommariva_46.json", _source)


def sommariva_47():
    return _read(this_dir / "sommariva_47.json", _source)


def sommariva_48():
    return _read(this_dir / "sommariva_48.json", _source)


def sommariva_49():
    return _read(this_dir / "sommariva_49.json", _source)


def sommariva_50():
    return _read(this_dir / "sommariva_50.json", _source)


def sommariva_51():
    return _read(this_dir / "sommariva_51.json", _source)


def sommariva_52():
    return _read(this_dir / "sommariva_52.json", _source)


def sommariva_53():
    return _read(this_dir / "sommariva_53.json", _source)


def sommariva_54():
    return _read(this_dir / "sommariva_54.json", _source)


def sommariva_55():
    return _read(this_dir / "sommariva_55.json", _source)


register(
    [
        sommariva_01,
        sommariva_02,
        sommariva_03,
        sommariva_04,
        sommariva_05,
        sommariva_06,
        sommariva_07,
        sommariva_08,
        sommariva_09,
        sommariva_10,
        sommariva_11,
        sommariva_12,
        sommariva_13,
        sommariva_14,
        sommariva_15,
        sommariva_16,
        sommariva_17,
        sommariva_18,
        sommariva_19,
        sommariva_20,
        sommariva_21,
        sommariva_22,
        sommariva_23,
        sommariva_24,
        sommariva_25,
        sommariva_26,
        sommariva_27,
        sommariva_28,
        sommariva_29,
        sommariva_30,
        sommariva_31,
        sommariva_32,
        sommariva_33,
        sommariva_34,
        sommariva_35,
        sommariva_36,
        sommariva_37,
        sommariva_38,
        sommariva_39,
        sommariva_40,
        sommariva_41,
        sommariva_42,
        sommariva_43,
        sommariva_44,
        sommariva_45,
        sommariva_46,
        sommariva_47,
        sommariva_48,
        sommariva_49,
        sommariva_50,
        sommariva_51,
        sommariva_52,
        sommariva_53,
        sommariva_54,
        sommariva_55,
    ]
)



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C2Scheme, register

source = article(
    authors=["W. Burnside"],
    title="An approximate quadrature formula",
    journal="Messenger of Math.",
    volume="37",
    year="1908",
    pages="166-167",
)


def burnside():
    d = {
        "d4_a0": [[frac(10, 49)], [sqrt(frac(7, 15))]],
        "d4_aa": [[frac(9, 196)], [sqrt(frac(7, 9))]],
    }
    return C2Scheme("Burnside", d, 5, source)


register([burnside])



import numpy as np

from ..c1 import gauss_legendre
from ..helpers import book
from ..tn._stroud import stroud_tn_5_1
from ._helpers import T3Scheme, register

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_t3_5_1():
    return stroud_tn_5_1(3)


def stroud_t3_7_1():
    degree = 7

    gl4 = gauss_legendre(4)
    r = (gl4.points + 1) / 2
    A = gl4.weights / 2

    # Generate Gauss formula for int_0^1 (1-s) * f(s) ds.
    # ```
    # k = np.arange(8)
    # moments = 1 / (k**2 + 3*k + 2)
    # alpha, beta = orthopy.c1.chebyshev(moments)
    # s, B = orthopy.c1.schemes.custom(alpha, beta, mode='numpy')
    # ```
    s = np.array(
        [
            5.710419611452533e-02,
            2.768430136381415e-01,
            5.835904323689318e-01,
            8.602401356562251e-01,
        ]
    )
    B = np.array(
        [
            1.355069134315012e-01,
            2.034645680102685e-01,
            1.298475476082247e-01,
            3.118097095000554e-02,
        ]
    )

    # Generate Gauss formula for int_0^1 (1-t)^2 * f(t) ds.
    # ```
    # k = np.arange(8)
    # moments = 2 / (k**3 + 6*k**2 + 11*k + 6)
    # alpha, beta = orthopy.c1.chebyshev(moments)
    # t, C = orthopy.c1.schemes.custom(alpha, beta, mode='numpy')
    # ```
    t = np.array(
        [
            4.850054944699245e-02,
            2.386007375518456e-01,
            5.170472951043522e-01,
            7.958514178967657e-01,
        ]
    )
    C = np.array(
        [
            1.108884156112685e-01,
            1.434587897992167e-01,
            6.863388717292915e-02,
            1.035224074991912e-02,
        ]
    )

    weights = np.array(
        [6 * A[i] * B[j] * C[k] for i in range(4) for j in range(4) for k in range(4)]
    )

    points = np.array(
        [
            [
                t[k],
                s[j] * (1 - t[k]),
                r[i] * (1 - s[j]) * (1 - t[k]),
                (1 - r[i]) * (1 - s[j]) * (1 - t[k]),
            ]
            for i in range(4)
            for j in range(4)
            for k in range(4)
        ]
    )
    points = np.ascontiguousarray(points.T)
    d = {"plain": [weights, points[0], points[1], points[2], points[3]]}
    return T3Scheme("Stroud T3 7-1", d, degree, source)


register([stroud_t3_5_1, stroud_t3_7_1])



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import E3rScheme, register

source = article(
    authors=["A.H. Stroud", "D. Secrest"],
    title="Approximate integration formulas for certain spherically symmetric regions",
    journal="Math. Comp.",
    volume="17",
    year="1963",
    pages="105-135",
    url="https://doi.org/10.1090/S0025-5718-1963-0161473-0",
)


def stroud_secrest_07():
    nu, xi = (sqrt(15 - p_m * 3 * sqrt(5)) for p_m in [+1, -1])
    A = frac(3, 5)
    B = frac(1, 30)

    d = {
        "zero3": [[A]],
        "symm_rs0_roll": [[B], [xi], [nu]],
    }
    return E3rScheme("Stroud-Secrest VII", d, 5, source)


def stroud_secrest_08():
    nu = sqrt(30)
    eta = sqrt(10)
    A = frac(3, 5)
    B = frac(2, 75)
    C = frac(3, 100)

    d = {
        "zero3": [[A]],
        "symm_r00": [[B], [nu]],
        "symm_rrr": [[C], [eta]],
    }
    return E3rScheme("Stroud-Secrest VIII", d, 5, source)


def stroud_secrest_09():
    eta = sqrt(10)
    xi, nu = (sqrt(15 - p_m * 5 * sqrt(5)) for p_m in [+1, -1])
    A = frac(3, 5)
    B = frac(1, 50)

    d = {
        "zero3": [[A]],
        "symm_rrr": [[B], [eta]],
        "symm_rs0_roll": [[B], [xi], [nu]],
    }
    return E3rScheme("Stroud-Secrest IX", d, 5, source)


def stroud_secrest_10():
    sqrt130 = sqrt(130)

    nu = sqrt((720 - 24 * sqrt130) / 11)
    xi = sqrt(288 + 24 * sqrt130)
    eta = sqrt((-216 + 24 * sqrt130) / 7)
    A = (5175 - 13 * sqrt130) / 8820
    B = (3870 + 283 * sqrt130) / 493920
    C = (3204 - 281 * sqrt130) / 197568
    # ERR in Stroud's book: 917568 vs. 197568
    D = (4239 + 373 * sqrt130) / 197568

    d = {
        "zero3": [[A]],
        "symm_r00": [[B], [nu]],
        "symm_rr0": [[C], [xi]],
        "symm_rrr": [[D], [eta]],
    }
    return E3rScheme("Stroud-Secrest X", d, 7, source)


def stroud_secrest_11():
    sqrt5 = sqrt(5)
    sqrt39 = sqrt(39)
    sqrt195 = sqrt(195)

    nu, xi = (
        sqrt(-50 + p_m * 10 * sqrt5 + 10 * sqrt39 - p_m * 2 * sqrt195)
        for p_m in [+1, -1]
    )
    eta = sqrt(36 + 4 * sqrt39)
    mu, lmbda = (
        sqrt(54 + p_m * 18 * sqrt5 + 6 * sqrt39 + p_m * 2 * sqrt195) for p_m in [+1, -1]
    )
    A = (1725 - 26 * sqrt39) / 2940
    B = (1065 + 171 * sqrt39) / 54880
    C = (297 - 47 * sqrt39) / 32928

    d = {
        "zero3": [[A]],
        "symm_rrr": [[C], [eta]],
        "symm_rs0_roll": [[B, C], [xi, lmbda], [nu, mu]],
    }
    return E3rScheme("Stroud-Secrest XI", d, 7, source, 1.487e-11)


register(
    [
        stroud_secrest_07,
        stroud_secrest_08,
        stroud_secrest_09,
        stroud_secrest_10,
        stroud_secrest_11,
    ]
)



import pathlib

from sympy import Rational as frac
from sympy import sqrt

from ...helpers import article
from .._helpers import T3Scheme, _read, register

source = article(
    authors=["Yu Jinyun"],
    title="Symmetyric Gaussian quadrature formulae for tetrahedronal regions",
    journal="Computer Methods in Applied Mechanics and Engineering",
    volume="43",
    year="1984",
    pages="349-353",
    url="https://doi.org/10.1016/0045-7825%2884%2990072-0",
)

this_dir = pathlib.Path(__file__).resolve().parent


def yu_2():
    degree = 2
    d = {"s31": [[frac(1, 4)], [frac(1, 4) - sqrt(5) * frac(1, 20)]]}
    return T3Scheme("Yu 1", d, degree, source)


def yu_3():
    degree = 3
    d = {"s4": [[-frac(4, 5)]], "s31": [[frac(9, 20)], [frac(1, 6)]]}
    return T3Scheme("Yu 2", d, degree, source)


def yu_4():
    return _read(this_dir / "yu_4.json", source)


def yu_5():
    return _read(this_dir / "yu_5.json", source)


def yu_6():
    return _read(this_dir / "yu_6.json", source)


register([yu_2, yu_3, yu_4, yu_5, yu_6])



import numpy as np
import sympy

from ..helpers import article
from ._helpers import S3Scheme, register

_source = article(
    authors=["I.P. Mysovskih"],
    title="On the construction of cubature formulas for the simplest regions",
    journal="Z. Vychisl. Mat. i. Mat. Fiz.",
    number="4",
    pages="3-14",
    year="1964",
)

frac = sympy.Rational
sqrt = np.vectorize(sympy.sqrt)
pi = sympy.pi
pm_ = np.array([+1, -1])


def mysovskih():
    sqrt17770 = sqrt(17770)
    r, s = sqrt((1715 - pm_ * 7 * sqrt17770) / 2817)
    t = sqrt(frac(7, 18))
    u = sqrt(frac(7, 27))

    B1, B2 = (2965 * sqrt17770 + pm_ * 227816) / 72030 / sqrt17770
    B3 = frac(324, 12005)
    B4 = frac(2187, 96040)

    d = {
        "symm_r00": [[B1, B2], [r, s]],
        "symm_rr0": [[B3], [t]],
        "symm_rrr": [[B4], [u]],
    }
    return S3Scheme("Mysovskih", d, 7, _source)


register([mysovskih])



import sympy

from ..helpers import article
from ._helpers import S2Scheme, register

_source = article(
    authors=["J. Albrecht", "L. Collatz"],
    title="Zur numerischen Auswertung mehrdimensionaler Integrale",
    journal="ZAMM",
    volume="38",
    number="1-2",
    year="1958",
    month="jan",
    pages="1–15",
    url="https://doi.org/10.1002/zamm.19580380102",
)

frac = sympy.Rational


def albrecht_collatz():
    # ERR Incorrectly stated in Stroud as sqrt(1/2) instead of 1/2
    d = {"d4_aa": [[frac(1, 4)], [frac(1, 2)]]}
    return S2Scheme("Albrecht-Collatz", d, 3, _source)


register([albrecht_collatz])



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["Sangwoo Heo", "Yuan Xu"],
    title="Constructing Fully Symmetric Cubature Formulae for the Sphere",
    journal="Mathematics of Computation",
    volume="70",
    number="233",
    month="jan",
    year="2001",
    pages="269-279",
    url="https://doi.org/10.1090/S0025-5718-00-01198-4",
)

this_dir = pathlib.Path(__file__).resolve().parent


def heo_xu_13():
    return _read(this_dir / "heo_xu_13.json", source)


def heo_xu_15():
    return _read(this_dir / "heo_xu_15.json", source)


def heo_xu_17():
    return _read(this_dir / "heo_xu_17.json", source)


def heo_xu_19a():
    return _read(this_dir / "heo_xu_19a.json", source)


def heo_xu_19b():
    return _read(this_dir / "heo_xu_19b.json", source)


def heo_xu_21a():
    return _read(this_dir / "heo_xu_21a.json", source)


def heo_xu_21b():
    return _read(this_dir / "heo_xu_21b.json", source)


def heo_xu_21c():
    return _read(this_dir / "heo_xu_21c.json", source)


def heo_xu_21d():
    return _read(this_dir / "heo_xu_21d.json", source)


def heo_xu_21e():
    return _read(this_dir / "heo_xu_21e.json", source)


def heo_xu_21f():
    return _read(this_dir / "heo_xu_21f.json", source)


def heo_xu_23a():
    return _read(this_dir / "heo_xu_23a.json", source)


def heo_xu_23b():
    return _read(this_dir / "heo_xu_23b.json", source)


def heo_xu_23c():
    return _read(this_dir / "heo_xu_23c.json", source)


def heo_xu_25a():
    return _read(this_dir / "heo_xu_25a.json", source)


def heo_xu_25b():
    return _read(this_dir / "heo_xu_25b.json", source)


def heo_xu_27a():
    return _read(this_dir / "heo_xu_27a.json", source)


def heo_xu_27b():
    return _read(this_dir / "heo_xu_27b.json", source)


def heo_xu_27c():
    return _read(this_dir / "heo_xu_27c.json", source)


def heo_xu_29():
    return _read(this_dir / "heo_xu_29.json", source)


def heo_xu_31():
    return _read(this_dir / "heo_xu_31.json", source)


def heo_xu_33():
    return _read(this_dir / "heo_xu_33.json", source)


def heo_xu_35():
    return _read(this_dir / "heo_xu_35.json", source)


def heo_xu_37():
    return _read(this_dir / "heo_xu_37.json", source)


def heo_xu_39a():
    return _read(this_dir / "heo_xu_39a.json", source)


def heo_xu_39b():
    return _read(this_dir / "heo_xu_39b.json", source)


register(
    [
        heo_xu_13,
        heo_xu_15,
        heo_xu_17,
        heo_xu_19a,
        heo_xu_19b,
        heo_xu_21a,
        heo_xu_21b,
        heo_xu_21c,
        heo_xu_21d,
        heo_xu_21e,
        heo_xu_21f,
        heo_xu_23a,
        heo_xu_23b,
        heo_xu_23c,
        heo_xu_25a,
        heo_xu_25b,
        heo_xu_27a,
        heo_xu_27b,
        heo_xu_27c,
        heo_xu_29,
        heo_xu_31,
        heo_xu_33,
        heo_xu_35,
        heo_xu_37,
        heo_xu_39a,
        heo_xu_39b,
    ]
)



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C2Scheme, register

source = article(
    authors=["J.C. Maxwell"],
    title="On Approximate Multiple Integration between Limits by Summation",
    journal="Cambridge Library Collection - Physical Sciences",
    pages="604-611",
    # publisher="Cambridge University Press",
    url="https://doi.org/10.1017/CBO9780511710377.061",
    note="In W. Niven (Ed.), The Scientific Papers of James Clerk Maxwell. First published in 1890.",
)


def maxwell():
    r = sqrt(frac(12, 35))
    s, t = (sqrt((93 + i * 3 * sqrt(186)) / 155) for i in [+1, -1])

    d = {
        "zero2": [[frac(1, 81)]],
        "d4_a0": [[frac(49, 324)], [r]],
        # ERR typo in Stroud: 648 vs 649
        "d4_ab": [[frac(31, 648)], [s], [t]],
    }
    return C2Scheme("Maxwell", d, 7, source)


register([maxwell])



from ._mcnamee_stenger import (
    mcnamee_stenger_3,
    mcnamee_stenger_5,
    mcnamee_stenger_7a,
    mcnamee_stenger_7b,
    mcnamee_stenger_9a,
    mcnamee_stenger_9b,
)
from ._stroud import (
    stroud_enr_3_1,
    stroud_enr_3_2,
    stroud_enr_5_1,
    stroud_enr_5_3,
    stroud_enr_5_4,
)
from ._stroud_secrest import (
    stroud_secrest_1,
    stroud_secrest_2,
    stroud_secrest_3,
    stroud_secrest_4,
)

__all__ = [
    "mcnamee_stenger_3",
    "mcnamee_stenger_5",
    "mcnamee_stenger_7a",
    "mcnamee_stenger_7b",
    "mcnamee_stenger_9a",
    "mcnamee_stenger_9b",
    "stroud_enr_3_1",
    "stroud_enr_3_2",
    "stroud_enr_5_1",
    "stroud_enr_5_3",
    "stroud_enr_5_4",
    "stroud_secrest_1",
    "stroud_secrest_2",
    "stroud_secrest_3",
    "stroud_secrest_4",
]



from typing import List

import sympy


def volume(n: int):
    return 2**n


def integrate_monomial(exponents: List[int], symbolic: bool = False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    exponents = list(exponents)
    n = len(exponents)

    if any(k % 2 == 1 for k in exponents):
        return 0

    def _recurrence(exponents):
        if all(k == 0 for k in exponents):
            return volume(n)

        # find first nonzero
        idx, k0 = next((i, k) for i, k in enumerate(exponents) if k > 0)
        k2 = exponents.copy()
        k2[idx] -= 2
        return _recurrence(k2) * frac(k0 - 1, k0 + 1)

    return _recurrence(exponents)



from typing import Callable

import ndim
import numpy as np

from .._exception import QuadpyError
from ..helpers import QuadratureScheme


class EnrScheme(QuadratureScheme):
    def __init__(self, name, dim, weights, points, degree, source, tol=1.0e-14):
        self.domain = f"Enr (n={dim})"
        self.dim = dim
        assert points.shape[0] == dim
        super().__init__(name, weights, points, degree, source, tol)

    def integrate(self, f: Callable, dot=np.dot):
        flt = np.vectorize(float)
        ref_vol = ndim.enr.volume(self.dim)

        x = flt(self.points)
        fx = np.asarray(f(x))

        assert len(x.shape) == 2
        if len(x.shape) == 2 and fx.shape[-1] != x.shape[1]:
            raise QuadpyError(
                f"Wrong return value shape {fx.shape}. Expected (:, {x.shape[1]})."
            )

        return ref_vol * dot(fx, flt(self.weights))



import numpy as np
from mpmath import mp
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, pm_roll, untangle
from ._helpers import U3Scheme, register

source = article(
    authors=["A.D. McLaren"],
    title="Optimal Numerical Integration on a Sphere",
    journal="Mathematics of Computation",
    volume="17",
    number="84",
    month="oct",
    year="1963",
    pages="361-383",
    url="https://doi.org/10.1090/S0025-5718-1963-0159418-2",
)


def mclaren_01():
    degree = 3
    data = {"a2": [frac(1, 12)]}
    return U3Scheme("McLaren 1", data, degree, source)


def mclaren_02():
    degree = 5
    r = frac(1, 2)
    s, t = ((sqrt(5) + pm_) / 4 for pm_ in [+1, -1])
    data = {"a1": [frac(1, 30)], "rst": [frac(1, 30), r, s, t]}
    return U3Scheme("McLaren 2", data, degree, source)


def mclaren_03():
    degree = 7

    # the positive roots of
    #  z^6 - z^4 + 0.2*z^2 - 1/105 = 0,
    # i.e., the square roots of the roots of
    #  z^3 - z^2 + 0.2*z^1 - 1/105 = 0,
    r2, s2, t2 = mp.polyroots([1, -1, frac(1, 5), -frac(1, 105)])
    r = sqrt(r2)
    s = sqrt(s2)
    t = sqrt(t2)

    data = {"rst_weird": [frac(1, 24), r, s, t]}
    return U3Scheme("McLaren 3", data, degree, source)


def mclaren_04():
    degree = 8

    # the positive roots of
    #  z^6 - z^4 + 5/21 * z^2 - 5/441 = 0,
    # i.e., the square roots of the roots of
    #  z^3 - z^2 + 5/21 * z^1 - 5/441 = 0,
    r2, s2, t2 = mp.polyroots([1, -1, frac(5, 21), -frac(5, 441)])
    r = sqrt(r2)
    s = sqrt(s2)
    t = sqrt(t2)

    data = {"a1": [frac(16, 600)], "rst_weird": [frac(21, 600), r, s, t]}
    return U3Scheme("McLaren 4", data, degree, source)


def mclaren_05():
    degree = 9

    r, s = (sqrt((5 + pm_ * sqrt(5)) / 10) for pm_ in [+1, -1])
    u, v = (sqrt((3 - pm_ * sqrt(5)) / 6) for pm_ in [+1, -1])

    B1 = frac(25, 840)
    B2 = frac(27, 840)

    data = {"rs0": [[B1, B2], [r, u], [s, v]], "a3": [B2]}
    return U3Scheme("McLaren 5", data, degree, source)


def mclaren_06():
    degree = 9

    r, s = (sqrt((5 + pm_ * sqrt(5)) / 10) for pm_ in [+1, -1])
    u = frac(1, 2)
    v, w = ((sqrt(5) + pm_) / 4 for pm_ in [+1, -1])

    B = frac(25, 1260)
    C = frac(32, 1260)

    # ERR Stroud is missing +- at the first r.
    data = {"rs0": [B, r, s], "a1": [C], "rst": [C, u, v, w]}
    return U3Scheme("McLaren 6", data, degree, source)


def mclaren_07():
    degree = 9

    r, s = (sqrt((3 - pm_ * sqrt(5)) / 6) for pm_ in [+1, -1])
    # ERR Stroud incorrectly gives sqrt(0.5)
    u = frac(1, 2)
    v, w = ((sqrt(5) + pm_) / 4 for pm_ in [+1, -1])

    B = -frac(9, 140)
    C = frac(16, 210)

    data = {"rs0": [B, r, s], "a3": [B], "a1": [C], "rst": [C, u, v, w]}
    return U3Scheme("McLaren 7", data, degree, source)


def mclaren_08():
    degree = 11

    u = sqrt(frac(1, 11))
    v = sqrt(frac(9, 11))

    B1 = frac(9216, 725760)
    B2 = frac(16384, 725760)
    B3 = frac(15309, 725760)
    B4 = frac(14641, 725760)

    data = {"a1": [B1], "a2": [B2], "a3": [B3], "llm2": [B4, u, v]}
    return U3Scheme("McLaren 8", data, degree, source)


def mclaren_09():
    degree = 11

    sqrt5 = sqrt(5)

    p, q = (sqrt((5 + pm_ * sqrt5) / 10) for pm_ in [+1, -1])
    r, s = (sqrt((3 - pm_ * sqrt5) / 6) for pm_ in [+1, -1])

    u = frac(1, 2)
    v, w = ((sqrt(5) + pm_) / 4 for pm_ in [+1, -1])

    B = frac(625, 27720)
    C = frac(243, 27720)
    D = frac(512, 27720)

    data = {"rs0": [[B, C], [p, r], [q, s]], "a3": [C], "a1": [D], "rst": [D, u, v, w]}
    return U3Scheme("McLaren 9", data, degree, source)


def mclaren_10():
    degree = 14

    r, s = (sqrt((5 - pm_ * sqrt(5)) / 10) for pm_ in [+1, -1])
    B = frac(125, 10080)
    C = frac(143, 10080)

    # The roots of
    #
    # 2556125 y^6 - 5112250 y^5 + 3578575 y^4 - 1043900 y^3
    #     + 115115 y^2 - 3562 y + 9 =0
    #
    # in decreasing order.
    y = [
        0.8318603575087328951583062165711519728388,
        0.5607526046766541293084396308069013490725,
        0.4118893592345073860321480490176804941547,
        0.1479981814629634692260834719469411619893,
        0.04473134613410273910111648293922113227845,
        0.002768150983039381173906148718103889666260,
    ]
    z = np.sqrt(y)

    u = (
        np.array([z[3] - z[2], z[1] - z[4], z[5] - z[1], z[2] - z[5], z[4] - z[3]])
        / 2
        / s
    )
    v = (
        np.array([z[4] + z[5], z[5] + z[3], z[2] + z[4], z[3] + z[1], z[1] + z[2]])
        / 2
        / s
    )
    w = (
        np.array([z[0] + z[1], z[0] + z[2], z[0] + z[3], z[0] + z[4], z[0] + z[5]])
        / 2
        / s
    )

    data = [
        (B, pm_roll([r, s, 0])),
        #
        (C, np.column_stack([+u, +v, +w])),
        (C, np.column_stack([+u, -v, -w])),
        (C, np.column_stack([-u, -v, +w])),
        (C, np.column_stack([-u, +v, -w])),
        #
        (C, np.column_stack([+v, +w, +u])),
        (C, np.column_stack([+v, -w, -u])),
        (C, np.column_stack([-v, -w, +u])),
        (C, np.column_stack([-v, +w, -u])),
        #
        (C, np.column_stack([+w, +u, +v])),
        (C, np.column_stack([+w, -u, -v])),
        (C, np.column_stack([-w, -u, +v])),
        (C, np.column_stack([-w, +u, -v])),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return U3Scheme(
        "McLaren 10", {"plain": np.vstack([weights, points])}, degree, source
    )


register(
    [
        mclaren_01,
        mclaren_02,
        mclaren_03,
        mclaren_04,
        mclaren_05,
        mclaren_06,
        mclaren_07,
        mclaren_08,
        mclaren_09,
        mclaren_10,
    ]
)



import pathlib

from ...helpers import techreport
from .._helpers import _read, register

source = techreport(
    authors=["M. Beckers", "A. Haegemans"],
    title="The construction of cubature formulae for the tetrahedron",
    institution="Dept. of Computer Science, K.U. Leuven",
    year="1990",
    note="Report TW 128",
    url="https://lirias.kuleuven.be/handle/123456789/132648",
)

this_dir = pathlib.Path(__file__).resolve().parent


def beckers_haegemans_8():
    return _read(this_dir / "beckers_haegemans_8.json", source)


def beckers_haegemans_9():
    return _read(this_dir / "beckers_haegemans_9.json", source)


register([beckers_haegemans_8, beckers_haegemans_9])



import numpy as np

from ..helpers import article, compute_dobrodeev, fsd, untangle
from ._helpers import CnScheme, integrate_monomial_over_ncube

_source = article(
    authors=["L.N. Dobrodeev"],
    title="Cubature rules with equal coefficients for integrating functions with respect to symmetric domains",
    journal="USSR Computational Mathematics and Mathematical Physics",
    volume="18",
    number="4",
    year="1978",
    pages="27-34",
    url="https://doi.org/10.1016/0041-5553%2878%2990064-2",
)


def dobrodeev_1978(n):
    assert 2 <= n <= 20

    dim_config = {
        2: ("II", 1, 1, 1),
        3: ("I", 3, 1, 1),
        4: ("I", 4, 1, 1),
        5: ("I", 5, 1, 1),
        6: ("I", 5, 1, 1),
        7: ("I", 6, 1, 2),
        8: ("I", 7, 1, 2),
        9: ("I", 6, 1, 2),
        10: ("I", 6, 1, 9),
        11: ("I", 6, 1, 10),
        12: ("I", 7, 1, 11),
        13: ("I", 8, 2, 2),
        14: ("I", 8, 1, 13),
        15: ("I", 9, 1, 14),
        16: ("I", 10, 2, 3),
        17: ("I", 10, 2, 3),
        18: ("I", 15, 3, 3),
        19: ("II", 9, 1, 18),
        20: ("I", 9, 1, 18),
    }

    ncube_limits = [[-1.0, 1.0]] * n
    I0 = integrate_monomial_over_ncube(ncube_limits, n * [0])
    I2 = integrate_monomial_over_ncube(ncube_limits, [2] + (n - 1) * [0])
    I22 = integrate_monomial_over_ncube(ncube_limits, [2, 2] + (n - 2) * [0])
    I4 = integrate_monomial_over_ncube(ncube_limits, [4] + (n - 1) * [0])

    pm_type, i, j, k = dim_config[n]

    G, a, b, c = compute_dobrodeev(n, I0, I2, I22, I4, pm_type, i, j, k, symbolic=True)

    data = [(G, fsd(n, (a, i))), (G, fsd(n, (b, j), (c, k)))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Dobrodeev 1978", n, weights, points, 5, _source, 8.350e-14)



from ..cn import ncube_points as cube_points
from ..cn import transform
from . import (
    _classical,
    _hammer_stroud,
    _hammer_wymore,
    _mustard_lyness_blatt,
    _sadowsky,
    _stroud,
    _stroud_1967,
    _tyler,
    _witherden_vincent,
)
from ._helpers import get_good_scheme, schemes
from ._product import product

__all__ = [
    "_classical",
    "_hammer_stroud",
    "_hammer_wymore",
    "_mustard_lyness_blatt",
    "_sadowsky",
    "_stroud",
    "_stroud_1967",
    "_tyler",
    "_witherden_vincent",
    #
    "product",
    #
    "transform",
    "cube_points",
    "schemes",
    "get_good_scheme",
]



import pathlib

from ...helpers import techreport
from .._helpers import _read, register

source = techreport(
    authors=["Noel J. Walkington"],
    title="Quadrature on simplices of arbitrary dimension",
    institution="CMU",
    year="2000",
    url="https://www.math.cmu.edu/~nw0z/publications/00-CNA-023/023abs/",
)

this_dir = pathlib.Path(__file__).resolve().parent


def walkington_p5():
    return _read(this_dir / "walkington_p5.json", source)


register([walkington_p5])



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["B. Vioreanu", "V. Rokhlin"],
    title="Spectra of Multiplication Operators as a Numerical Tool",
    year="2014",
    journal="SIAM J. Sci. Comput.",
    volume="36",
    number="1",
    url="https://doi.org/10.1137/110860082",
)
# http://www.cs.yale.edu/publications/techreports/tr1443.pdf

# Data adapted from modepy
# https://github.com/inducer/modepy/blob/master/modepy/quadrature/vr_quad_data_tri.py

this_dir = pathlib.Path(__file__).resolve().parent


def vioreanu_rokhlin_0():
    return _read(this_dir / "vr00.json", source)


def vioreanu_rokhlin_1():
    return _read(this_dir / "vr01.json", source)


def vioreanu_rokhlin_2():
    return _read(this_dir / "vr02.json", source)


def vioreanu_rokhlin_3():
    return _read(this_dir / "vr03.json", source)


def vioreanu_rokhlin_4():
    return _read(this_dir / "vr04.json", source)


def vioreanu_rokhlin_5():
    return _read(this_dir / "vr05.json", source)


def vioreanu_rokhlin_6():
    return _read(this_dir / "vr06.json", source)


def vioreanu_rokhlin_7():
    return _read(this_dir / "vr07.json", source)


def vioreanu_rokhlin_8():
    return _read(this_dir / "vr08.json", source)


def vioreanu_rokhlin_9():
    return _read(this_dir / "vr09.json", source)


register(
    [
        vioreanu_rokhlin_0,
        vioreanu_rokhlin_1,
        vioreanu_rokhlin_2,
        vioreanu_rokhlin_3,
        vioreanu_rokhlin_4,
        vioreanu_rokhlin_5,
        vioreanu_rokhlin_6,
        vioreanu_rokhlin_7,
        vioreanu_rokhlin_8,
        vioreanu_rokhlin_9,
    ]
)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["Hong Xiao", "Zydrunas Gimbutas"],
    title="A numerical algorithm for the construction of efficient quadrature rules in two and higher dimensions",
    journal="Computers & Mathematics with Applications",
    volume="59",
    number="2",
    month="jan",
    year="2010",
    pages="663–676",
    url="https://doi.org/10.1016/j.camwa.2009.10.027",
)
# Data extracted from
# https://people.sc.fsu.edu/~jburkardt/f_src/triangle_symq_rule/triangle_symq_rule.f90

this_dir = pathlib.Path(__file__).resolve().parent


def xiao_gimbutas_01():
    return _read(this_dir / "xg01.json", source)


def xiao_gimbutas_02():
    return _read(this_dir / "xg02.json", source)


def xiao_gimbutas_03():
    return _read(this_dir / "xg03.json", source)


def xiao_gimbutas_04():
    return _read(this_dir / "xg04.json", source)


def xiao_gimbutas_05():
    return _read(this_dir / "xg05.json", source)


def xiao_gimbutas_06():
    return _read(this_dir / "xg06.json", source)


def xiao_gimbutas_07():
    return _read(this_dir / "xg07.json", source)


def xiao_gimbutas_08():
    return _read(this_dir / "xg08.json", source)


def xiao_gimbutas_09():
    return _read(this_dir / "xg09.json", source)


def xiao_gimbutas_10():
    return _read(this_dir / "xg10.json", source)


def xiao_gimbutas_11():
    return _read(this_dir / "xg11.json", source)


def xiao_gimbutas_12():
    return _read(this_dir / "xg12.json", source)


def xiao_gimbutas_13():
    return _read(this_dir / "xg13.json", source)


def xiao_gimbutas_14():
    return _read(this_dir / "xg14.json", source)


def xiao_gimbutas_15():
    return _read(this_dir / "xg15.json", source)


def xiao_gimbutas_16():
    return _read(this_dir / "xg16.json", source)


def xiao_gimbutas_17():
    return _read(this_dir / "xg17.json", source)


def xiao_gimbutas_18():
    return _read(this_dir / "xg18.json", source)


def xiao_gimbutas_19():
    return _read(this_dir / "xg19.json", source)


def xiao_gimbutas_20():
    return _read(this_dir / "xg20.json", source)


def xiao_gimbutas_21():
    return _read(this_dir / "xg21.json", source)


def xiao_gimbutas_22():
    return _read(this_dir / "xg22.json", source)


def xiao_gimbutas_23():
    return _read(this_dir / "xg23.json", source)


def xiao_gimbutas_24():
    return _read(this_dir / "xg24.json", source)


def xiao_gimbutas_25():
    return _read(this_dir / "xg25.json", source)


def xiao_gimbutas_26():
    return _read(this_dir / "xg26.json", source)


def xiao_gimbutas_27():
    return _read(this_dir / "xg27.json", source)


def xiao_gimbutas_28():
    return _read(this_dir / "xg28.json", source)


def xiao_gimbutas_29():
    return _read(this_dir / "xg29.json", source)


def xiao_gimbutas_30():
    return _read(this_dir / "xg30.json", source)


def xiao_gimbutas_31():
    return _read(this_dir / "xg31.json", source)


def xiao_gimbutas_32():
    return _read(this_dir / "xg32.json", source)


def xiao_gimbutas_33():
    return _read(this_dir / "xg33.json", source)


def xiao_gimbutas_34():
    return _read(this_dir / "xg34.json", source)


def xiao_gimbutas_35():
    return _read(this_dir / "xg35.json", source)


def xiao_gimbutas_36():
    return _read(this_dir / "xg36.json", source)


def xiao_gimbutas_37():
    return _read(this_dir / "xg37.json", source)


def xiao_gimbutas_38():
    return _read(this_dir / "xg38.json", source)


def xiao_gimbutas_39():
    return _read(this_dir / "xg39.json", source)


def xiao_gimbutas_40():
    return _read(this_dir / "xg40.json", source)


def xiao_gimbutas_41():
    return _read(this_dir / "xg41.json", source)


def xiao_gimbutas_42():
    return _read(this_dir / "xg42.json", source)


def xiao_gimbutas_43():
    return _read(this_dir / "xg43.json", source)


def xiao_gimbutas_44():
    return _read(this_dir / "xg44.json", source)


def xiao_gimbutas_45():
    return _read(this_dir / "xg45.json", source)


def xiao_gimbutas_46():
    return _read(this_dir / "xg46.json", source)


def xiao_gimbutas_47():
    return _read(this_dir / "xg47.json", source)


def xiao_gimbutas_48():
    return _read(this_dir / "xg48.json", source)


def xiao_gimbutas_49():
    return _read(this_dir / "xg49.json", source)


def xiao_gimbutas_50():
    return _read(this_dir / "xg50.json", source)


register(
    [
        xiao_gimbutas_01,
        xiao_gimbutas_02,
        xiao_gimbutas_03,
        xiao_gimbutas_04,
        xiao_gimbutas_05,
        xiao_gimbutas_06,
        xiao_gimbutas_07,
        xiao_gimbutas_08,
        xiao_gimbutas_09,
        xiao_gimbutas_10,
        xiao_gimbutas_11,
        xiao_gimbutas_12,
        xiao_gimbutas_13,
        xiao_gimbutas_14,
        xiao_gimbutas_15,
        xiao_gimbutas_16,
        xiao_gimbutas_17,
        xiao_gimbutas_18,
        xiao_gimbutas_19,
        xiao_gimbutas_20,
        xiao_gimbutas_21,
        xiao_gimbutas_22,
        xiao_gimbutas_23,
        xiao_gimbutas_24,
        xiao_gimbutas_25,
        xiao_gimbutas_26,
        xiao_gimbutas_27,
        xiao_gimbutas_28,
        xiao_gimbutas_29,
        xiao_gimbutas_30,
        xiao_gimbutas_31,
        xiao_gimbutas_32,
        xiao_gimbutas_33,
        xiao_gimbutas_34,
        xiao_gimbutas_35,
        xiao_gimbutas_36,
        xiao_gimbutas_37,
        xiao_gimbutas_38,
        xiao_gimbutas_39,
        xiao_gimbutas_40,
        xiao_gimbutas_41,
        xiao_gimbutas_42,
        xiao_gimbutas_43,
        xiao_gimbutas_44,
        xiao_gimbutas_45,
        xiao_gimbutas_46,
        xiao_gimbutas_47,
        xiao_gimbutas_48,
        xiao_gimbutas_49,
        xiao_gimbutas_50,
    ]
)



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, get_nsimplex_points
from ._helpers import UnScheme

source = article(
    authors=["I.P. Mysovskikh"],
    title="The approximation of multiple integrals by using interpolatory cubature formulae",
    journal="Proceedings of a Symposium on Quantitative Approximation Held in Bonn, West Germany, August 20–24, 1979",
    year="1980",
    pages="217-243",
    url="https://doi.org/10.1016/B978-0-12-213650-4.50025-8",
)


def mysovskikh_1(n):
    points = get_nsimplex_points(n, sqrt, frac)
    weights = np.full(n + 1, frac(1, n + 1))
    return UnScheme("Mysovskikh 1", n, weights, points, 2, source)


def mysovskikh_2(n):
    A = frac((7 - n) * n, 2 * (n + 1) ** 2 * (n + 2))
    B = frac(2 * (n - 1) ** 2, n * (n + 1) ** 2 * (n + 2))

    a = get_nsimplex_points(n, sqrt, frac)
    b = np.array(
        [
            sqrt(frac(n, 2 * (n - 1))) * (a[k] + a[l])
            for k in range(n + 1)
            for l in range(k)
        ]
    )

    points = np.concatenate([a, -a, b, -b])
    weights = np.concatenate(
        [
            np.full(len(a), A),
            np.full(len(a), A),
            np.full(len(b), B),
            np.full(len(b), B),
        ]
    )
    return UnScheme("Mysovskikh 2", n, weights, points, 5, source)



from . import _classical, _ditkin, _hammer_stroud, _mysovskih, _stroud
from ._helpers import get_good_scheme, schemes

__all__ = [
    "_classical",
    "_ditkin",
    "_hammer_stroud",
    "_mysovskih",
    "_stroud",
    #
    "schemes",
    "get_good_scheme",
]



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, expand_symmetries, fsd, untangle, z
from ._helpers import CnScheme

_source = article(
    authors=["Preston C. Hammer", "Arthur H. Stroud"],
    title="Numerical Evaluation of Multiple Integrals II",
    journal="Math. Comp.",
    volume="12",
    year="1958",
    pages="272-280",
    url="https://doi.org/10.1090/S0025-5718-1958-0102176-6",
)


def hammer_stroud_1n(n):
    d = {"a0": [[frac(1, 2 * n)], [sqrt(frac(n, 3))]]}
    points, weights = expand_symmetries(d, n)
    return CnScheme("Hammer-Stroud 1n", n, weights, points, 3, _source, 5.863e-14)


def hammer_stroud_2n(n):
    r = sqrt(frac(3, 5))
    data = [
        (frac(25 * n ** 2 - 115 * n + 162, 162), z(n)),
        (frac(70 - 25 * n, 162), fsd(n, (r, 1))),
        (frac(25, 324), fsd(n, (r, 2))),
    ]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Hammer-Stroud 2n", n, weights, points, 5, _source, 3.820e-14)



from __future__ import annotations

import warnings

import numpy as np


class QuadratureScheme:
    def __init__(
        self,
        name: str,
        weights,
        points,
        degree: int,
        source,
        tol: float = 1.0e-14,
        comments: list[str] | None = None,
    ):
        if tol > 1.0e-12:
            warnings.warn(f"{name} ({self.domain}) has low precision ({tol:.3e}).")

        self.test_tolerance = tol
        self.name = name
        self.degree = degree
        self.source = source
        self.comments = [] if comments is None else comments

        # TODO enable contiguity check
        # assert points.flags["C_CONTIGUOUS"], name
        # assert weights.flags["C_CONTIGUOUS"], name

        # assert weights.shape[0] == points.shape[1], (
        #     f"Shape mismatch for {name}: "
        #     f"weights.shape = {weights.shape}, points.shape = {points.shape}"
        # )

        if weights.dtype == np.float64:
            self.weights = weights
        else:
            assert weights.dtype in [np.dtype("O"), np.int_]
            self.weights = weights.astype(np.float64)
            self.weights_symbolic = weights

        if points.dtype == np.float64:
            self.points = points
        else:
            assert points.dtype in [np.dtype("O"), np.int_]
            self.points = points.astype(np.float64)
            self.points_symbolic = points

    def savefig(self, filename, *args, **kwargs):
        from matplotlib import pyplot as plt

        self.plot(*args, **kwargs)
        # mpl keeps a hidden background patch that renders bbox_inches ineffective.
        # keep an eye out for https://stackoverflow.com/q/61712551/353337
        plt.savefig(filename, transparent=True, bbox_inches="tight", pad_inches=0)

    def show(self, *args, **kwargs):
        from matplotlib import pyplot as plt

        self.plot(*args, **kwargs)
        plt.show()

    def __str__(self):
        message = f"<quadrature scheme for {self.domain}>"
        message += f"\n  name:                 {self.name}"
        try:
            source = self.source
        except AttributeError:
            pass
        else:
            if source is not None:
                message += f"\n  source:               {source.title}"
                if source.authors:
                    authors = ", ".join(source.authors)
                    message += f"\n                        {authors}"

                string = []
                try:
                    if source.journal:
                        string.append(source.journal)
                except AttributeError:
                    pass
                try:
                    if source.volume:
                        string.append(f"vol. {source.volume}")
                except AttributeError:
                    pass
                try:
                    if source.number:
                        string.append(f"no. {source.number}")
                except AttributeError:
                    pass
                try:
                    if source.pages:
                        string.append(f"pp. {source.pages}")
                except AttributeError:
                    pass
                try:
                    if source.year:
                        string.append(source.year)
                except AttributeError:
                    pass

                if string:
                    s = ", ".join(string)
                    message += f"\n                        {s}"
                if source.url:
                    message += f"\n                        {source.url}"

        message += f"\n  degree:               {self.degree}"
        message += f"\n  num points/weights:   {len(self.weights)}"

        max_weight = np.max(np.abs(self.weights))
        min_weight = np.min(np.abs(self.weights))
        weights_ratio = max_weight / min_weight
        message += f"\n  max/min weight ratio: {weights_ratio:.3e}"

        try:
            message += f"\n  test tolerance:       {self.test_tolerance}"
        except AttributeError:
            pass

        try:
            pi = self.points_inside()
        except AttributeError:
            pass
        else:
            if pi.all():
                point_position = "strictly inside"
            elif self.points_inside_or_boundary().all():
                point_position = "inside + boundary"
            else:
                point_position = "outside"
            message += f"\n  point position:       {point_position}"

        weights_positive = all(self.weights > 0)
        message += f"\n  all weights positive: {weights_positive}"

        try:
            comments = self.comments
        except AttributeError:
            pass
        else:
            if len(comments) > 0:
                message += "\n  comments:"
                for comment in comments:
                    message += f"\n    {comment}"

        return message
